############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
# Python  v3.7.4
#
# Michael Würtenberger
# (c) 2019
#
# Licence APL2.0
#
###########################################################
# standard libraries
import os
from pathlib import Path
# external packages
import PyQt5
from astropy.io import fits
# local import
from mw4.base.tpool import Worker


class Satellite(object):
    """
    the Satellite window class handles the main menu as well as the show and no show part of
    any other window. all necessary processing for functions of that gui will be linked
    to this class. therefore window classes will have a threadpool for managing async
    processing if needed.
    """

    def __init__(self):
        self. satellites = list()
        self.satellite = None
        self.satelliteTLE = {}

        self.satelliteSourceDropDown = {
            'Active': 'http://www.celestrak.com/NORAD/elements/active.txt',
            'Space Stations': 'http://www.celestrak.com/NORAD/elements/stations.txt',
            '100 brightest': 'http://www.celestrak.com/NORAD/elements/visual.txt',
            'NOAA': 'http://www.celestrak.com/NORAD/elements/noaa.txt',
            'GEOS': 'http://www.celestrak.com/NORAD/elements/goes.txt',
            'Weather': 'http://www.celestrak.com/NORAD/elements/weather.txt',
            'Earth Resources': 'http://celestrak.com/NORAD/elements/resource.txt',
            'TDRSS TRacking and Data Relay': 'http://celestrak.com/NORAD/elements/tdrss.txt',
            'ARGOS': 'http://celestrak.com/NORAD/elements/argos.txt',
            'Amateur Radio': 'http://www.celestrak.com/NORAD/elements/amateur.txt',
            'Space & Earth Science': 'http://celestrak.com/NORAD/elements/science.txt',
            'Engineering': 'http://celestrak.com/NORAD/elements/engineering.txt',
            'Last 30 days launch': 'http://www.celestrak.com/NORAD/elements/tle-new.txt',
        }

        self.ui.listSatelliteNames.itemPressed.connect(self.extractSatelliteData)
        self.ui.startSatelliteTracking.clicked.connect(self.startTrack)
        self.ui.stopSatelliteTracking.clicked.connect(self.stopTrack)
        self.ui.satelliteSource.currentIndexChanged.connect(self.loadSatelliteSource)

        self.app.mount.signals.calcTLEdone.connect(self.enableTrack)
        self.app.mount.signals.getTLEdone.connect(self.prepare)
        self.ui.isOnline.stateChanged.connect(self.loadSatelliteSource)

        self.app.update3s.connect(self.updateSatelliteData)

    def initConfig(self):
        """
        initConfig read the key out of the configuration dict and stores it to the gui
        elements. if some initialisations have to be proceeded with the loaded persistent
        data, they will be launched as well in this method.

        :return: True for test purpose
        """

        config = self.app.config['mainW']
        self.setupSatelliteSourceGui()
        self.loadSatelliteSource()
        return True

    def storeConfig(self):
        """
        storeConfig writes the keys to the configuration dict and stores. if some
        saving has to be proceeded to persistent data, they will be launched as
        well in this method.

        :return: True for test purpose
        """
        config = self.app.config['mainW']
        return True

    def setupIcons(self):
        """
        setupIcons add icon from standard library to certain buttons for improving the
        gui of the app.

        :return:    True if success for test
        """

        return True

    def setupSatelliteSourceGui(self):
        """
        setupSatelliteSourceGui handles the dropdown list for the satellite data online
        sources. therefore we add the necessary entries to populate the list.

        :return: success for test
        """

        self.ui.satelliteSource.clear()
        self.ui.satelliteSource.setView(PyQt5.QtWidgets.QListView())
        for name in self.satelliteSourceDropDown.keys():
            self.ui.satelliteSource.addItem(name)

        return True

    def prepare(self, tleParams=None):
        """

        :param tleParams:
        :return: True for test purpose
        """

        if not tleParams:
            return False

        self.extractSatelliteData(None, satName=tleParams.name)
        return True

    def setupSatelliteGui(self):
        """
        setupSatelliteGui clears the list view of satellite names deriving from the selected
        source file on disk. after that it populated the list with actual data.

        :return: success for test
        """

        self.ui.listSatelliteNames.clear()
        for name, _ in self.satellites.items():
            if not isinstance(name, str):
                continue
            entryName = f'{self.satellites[name].model.satnum:5d} - {name}'
            self.ui.listSatelliteNames.addItem(entryName)
        self.ui.listSatelliteNames.sortItems()
        self.ui.listSatelliteNames.update()

        return True

    def loadTLEData(self, source=''):
        """
        loadTLEData load the two line elements from the source file and stores is separately
        in a dictionary, because we need that data later for transfer it to the mount
        computer. unfortunately the loader from skyfield does not store the original TLE
        data, but only parses it and throws the original away.

        :param source:
        :return: success
        """

        if not source:
            return False

        fileName = os.path.basename(source)
        dirPath = self.app.mwGlob['dataDir']
        filePath = f'{dirPath}/{fileName}'

        if not os.path.isfile(filePath):
            return False

        self.satelliteTLE = {}
        with open(filePath, mode='r') as tleFile:
            while True:
                l0 = tleFile.readline()
                l1 = tleFile.readline()
                l2 = tleFile.readline()

                if not l0:
                    break
                self.satelliteTLE[l0.strip()] = {'line0': l0.strip('\n'),
                                                 'line1': l1.strip('\n'),
                                                 'line2': l2.strip('\n'),
                                                 }
        return True

    def loadSatelliteSourceWorker(self):
        """
        loadSatelliteSourceWorker selects from a drop down list of possible satellite
        data sources on the web and once selected downloads the data. depending of the
        setting of reload is true setting, it takes an already loaded file from local disk.
        after loading or opening the source file, it updates the satellite list in the list
        view widget for the selection of satellites.

        :return: success
        """

        key = self.ui.satelliteSource.currentText()

        if key not in self.satelliteSourceDropDown:
            return False

        source = self.satelliteSourceDropDown[key]
        reload = self.ui.isOnline.isChecked()
        self.satellites = self.app.loader.tle(source, reload=reload)

        suc = self.loadTLEData(source)
        if not suc:
            return False

        return True

    def loadSatelliteSource(self):
        """
        loadSatelliteSource selects from a drop down list of possible satellite data sources
        on the web and once selected downloads the data. depending of the setting of reload
        is true setting, it takes an already loaded file from local disk.
        after loading or opening the source file, it updates the satellite list in the list
        view widget for the selection of satellites.

        :return: success
        """

        worker = Worker(self.loadSatelliteSourceWorker)
        worker.signals.result.connect(self.setupSatelliteGui)
        self.threadPool.start(worker)

        return True

    def updateSatelliteData(self):
        """
        updateSatelliteData calculates the actual satellite orbits, subpoint etc. and
        updates the data in the gui. in addition when satellite window is open it signals
        this update data as well for matplotlib drawings in satellite window.
        this method is called cyclic every 3 seconds for updates

        :return: success
        """

        if self.satellite is None:
            self.ui.startSatelliteTracking.setEnabled(False)
            return False

        # check if calculation is necessary to optimize cpu time
        # get index for satellite tab
        tabWidget = self.ui.mainTabWidget.findChild(PyQt5.QtWidgets.QWidget, 'Satellite')
        tabIndex = self.ui.mainTabWidget.indexOf(tabWidget)
        satTabVisible = self.ui.mainTabWidget.currentIndex() == tabIndex

        if not self.app.satelliteW and not satTabVisible:
            return False

        now = self.app.mount.obsSite.ts.now()
        observe = self.satellite.at(now)
        sett = self.app.mount.sett

        subpoint = observe.subpoint()
        difference = self.satellite - self.app.mount.obsSite.location

        ra, dec, _ = difference.at(now).radec()
        ra = ra.hours
        dec = dec.degrees

        self.ui.satRa.setText(f'{ra:3.2f}')
        self.ui.satDec.setText(f'{dec:3.2f}')

        lat = subpoint.latitude.degrees
        lon = subpoint.longitude.degrees
        self.ui.satLatitude.setText(f'{lat:3.2f}')
        self.ui.satLongitude.setText(f'{lon:3.2f}')

        hasPressure = (sett.refractionPress is not None)
        hasTemperature = (sett.refractionTemp is not None)

        if hasPressure and hasTemperature:
            altaz = difference.at(now).altaz(temperature_C=sett.refractionPress,
                                             pressure_mbar=sett.refractionTemp)
        else:
            altaz = difference.at(now).altaz()

        alt, az, _ = altaz
        alt = alt.degrees
        az = az.degrees

        self.ui.satAltitude.setText(f'{alt:3.2f}')
        self.ui.satAzimuth.setText(f'{az:3.2f}')

        if not self.app.satelliteW:
            return True

        self.app.satelliteW.signals.update.emit(observe, subpoint, altaz)
        return True

    def programTLEToMount(self):
        """
        programTLEToMount get the satellite data and programs this TLE data into the mount.
        after programming the parameters it forces the mount to calculate the satellite
        orbits immediately

        :return: success
        """

        if not self.app.mount.mountUp:
            self.app.message.emit('Mount is not online', 2)
            return False

        satellite = self.app.mount.satellite
        if satellite.tleParams.name == self.satellite.name:
            self.app.message.emit(f'Actual satellite is  [{self.satellite.name}]', 0)
        else:
            self.app.message.emit(f'Programming [{self.satellite.name}] to mount', 0)
            data = self.satelliteTLE[self.satellite.name]
            suc = satellite.setTLE(line0=data['line0'],
                                   line1=data['line1'],
                                   line2=data['line2'])
            if not suc:
                self.app.message.emit('Error program TLE', 2)
                return False

        return True

    def calcTLEParams(self):
        """
        calcTLEParams is called cyclic to update the orbit parameters in the mount computer

        :return: success
        """

        if self.satellite is None:
            self.ui.startSatelliteTracking.setEnabled(False)
            self.ui.stopSatelliteTracking.setEnabled(False)
            return False

        # starting thread for calculation of parameters
        self.app.mount.calcTLE()

        return True

    def extractSatelliteData(self, widget, satName=''):
        """
        extractSatelliteData is called when a satellite is selected via mouse click in the
        list menu. it collects the data and writes basic stuff to the gui.
        for speeding up, is calls updateSatelliteData immediately to get the actual data
        pushed to the gui and not waiting for the cyclic task.
        depending on the age of the satellite data is colors the frame

        :param widget:
        :param satName:
        :return: success
        """

        if not isinstance(satName, str):
            return False

        if not satName:
            satName = self.ui.listSatelliteNames.currentItem().text()[8:]

        if satName not in self.satellites:
            return False

        for index in range(self.ui.listSatelliteNames.model().rowCount()):
            if not self.ui.listSatelliteNames.item(index).text()[8:] == satName:
                continue
            item = self.ui.listSatelliteNames.item(index)
            item.setSelected(True)
            break

        position = PyQt5.QtWidgets.QAbstractItemView.EnsureVisible
        self.ui.listSatelliteNames.scrollToItem(item, position)
        self.satellite = self.satellites[satName]

        self.ui.satelliteName.setText(self.satellite.name)
        epochText = self.satellite.epoch.utc_strftime('%Y-%m-%d, %H:%M:%S')
        self.ui.satelliteEpoch.setText(epochText)

        now = self.app.mount.obsSite.ts.now()
        days = now - self.satellite.epoch
        self.ui.satelliteDataAge.setText(f'{days:2.2f}')

        if days > 10:
            self.changeStyleDynamic(self.ui.satelliteDataAge, 'color', 'red')
        elif 3 < days < 10:
            self.changeStyleDynamic(self.ui.satelliteDataAge, 'color', 'yellow')
        else:
            self.changeStyleDynamic(self.ui.satelliteDataAge, 'color', '')

        self.ui.satelliteNumber.setText(f'{self.satellite.model.satnum:5d}')

        self.ui.stopSatelliteTracking.setEnabled(False)
        self.ui.startSatelliteTracking.setEnabled(False)
        self.ui.satTransitStartUTC.setText('-')
        self.ui.satTransitEndUTC.setText('-')
        self.ui.satNeedFlip.setText('-')

        self.updateSatelliteData()
        self.programTLEToMount()
        self.calcTLEParams()

        if not self.app.satelliteW:
            return False

        self.app.satelliteW.signals.show.emit(self.satellite)
        return True

    def enableTrack(self, tleParams=None):
        """

        :return: success for test purpose
        """
        if not tleParams:
            return False

        if tleParams.jdStart is not None:
            time = self.app.mount.obsSite.ts.tt_jd(tleParams.jdStart)
            self.ui.satTransitStartUTC.setText(time.utc_strftime('%Y-%m-%d  %H:%M:%S'))
        else:
            self.ui.satTransitStartUTC.setText('No transit')

        if tleParams.jdEnd is not None:
            time = self.app.mount.obsSite.ts.tt_jd(tleParams.jdEnd)
            self.ui.satTransitEndUTC.setText(time.utc_strftime('%Y-%m-%d  %H:%M:%S'))
        else:
            self.ui.satTransitEndUTC.setText('No transit')

        if tleParams.flip:
            self.ui.satNeedFlip.setText('YES')
        else:
            self.ui.satNeedFlip.setText('NO')

        if tleParams.message is not None:
            self.app.message.emit(tleParams.message, 0)

        if tleParams.altitude is not None:
            self.ui.stopSatelliteTracking.setEnabled(True)
            self.ui.startSatelliteTracking.setEnabled(True)
        else:
            self.ui.stopSatelliteTracking.setEnabled(False)
            self.ui.startSatelliteTracking.setEnabled(False)

        return True

    def startTrack(self):
        """
        startTrack un parks the mount if the mount has this state, because tracking could not
        start with mount parked. if unparked, slewing is initiated.


        :return: success
        """

        if not self.app.mount.mountUp:
            self.app.message.emit('Mount is not online', 2)
            return False

        # if mount is currently parked, we unpark it
        if self.app.mount.obsSite.status == 5:
            suc = self.app.mount.obsSite.unpark()
            if not suc:
                self.app.message.emit('Cannot unpark mount', 2)
            else:
                self.app.message.emit('Mount unparked', 0)

        suc, message = self.app.mount.satellite.slewTLE()
        if not suc:
            self.app.message.emit(message, 2)
            return False

        self.app.message.emit(message, 0)
        return True

    def stopTrack(self):
        """
        stopTrack just sends the command stop tracking. this is also valid for satellite
        tracking

        :return: success
        """

        if not self.app.mount.mountUp:
            self.app.message.emit('Mount is not online', 2)
            return False

        suc = self.app.mount.obsSite.stopTracking()
        if not suc:
            self.app.message.emit('Cannot stop tracking', 2)
        else:
            self.app.message.emit('Stopped tracking', 0)

        return suc



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
# Python  v3.7.4
#
# Michael Würtenberger
# (c) 2019
#
# Licence APL2.0
#
###########################################################
# standard libraries
import logging
import re
import time
# external packages
import PyQt5
import requests
# local imports


class KMRelay(PyQt5.QtCore.QObject):
    """
    The class KMRelay inherits all information and handling of KMtronic relay board
    attributes of the connected board and provides the abstracted interface.

        >>> fw = KMRelay(
        >>>                 host=host
        >>>                 user=user
        >>>                 password=password
        >>>              )
    """

    __all__ = ['KMRelay',
               'startTimers',
               'stopTimers',
               'cyclePolling',
               'pulse',
               'switch',
               'set',
               ]

    version = '0.100.0'
    logger = logging.getLogger(__name__)

    # polling cycle for relay box
    CYCLE_POLLING = 1000
    # default port for KMTronic Relay
    DEFAULT_PORT = 80
    # timeout for requests
    TIMEOUT = 0.5
    # width for pulse
    PULSEWIDTH = 0.5
    # signal if correct status received and decoded
    statusReady = PyQt5.QtCore.pyqtSignal()

    def __init__(self,
                 host=None,
                 user=None,
                 password=None,
                 ):
        super().__init__()

        self.host = host
        self.mutexPoll = PyQt5.QtCore.QMutex()
        self.user = user
        self.password = password
        self.status = [0] * 8

        self.timerTask = PyQt5.QtCore.QTimer()
        self.timerTask.setSingleShot(False)
        self.timerTask.timeout.connect(self.cyclePolling)

    @property
    def host(self):
        return self._host

    def checkFormat(self, value):
        # checking format
        if not value:
            self.logger.error(f'Wrong host value: {value}')
            return None
        if not isinstance(value, (tuple, str)):
            self.logger.error(f'Wrong host value: {value}')
            return None
        # now we got the right format
        if isinstance(value, str):
            value = (value, self.DEFAULT_PORT)
        return value

    @host.setter
    def host(self, value):
        value = self.checkFormat(value)
        self._host = value

    @property
    def user(self):
        return self._user

    @user.setter
    def user(self, value):
        self._user = value

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, value):
        self._password = value

    def startTimers(self):
        """
        startTimers enables the cyclic timers for polling necessary relay data.

        :return: success
        """

        if self._host is None:
            return False
        if not self._host[0]:
            return False
        if not self._host[1]:
            return False

        self.timerTask.start(self.CYCLE_POLLING)

        return True

    def stopTimers(self):
        """
        stopTimers disables the cyclic timers for polling necessary relay data.

        :return: True for test purpose
        """
        self.timerTask.stop()
        return True

    def debugOutput(self, result=None):
        """
        debugOutput writes a nicely formed output for diagnosis in relay environment

        :param result: value from requests get commend
        :return: True for test purpose
        """

        if result is None:
            self.logger.debug('No valid result')
            return False

        text = result.text.replace('\r\n', ', ')
        reason = result.reason
        status = result.status_code
        url = result.url
        elapsed = result.elapsed

        self.logger.debug(f'Result: {url}, {reason}, {status}, {elapsed}, {text}')

        return True

    def getRelay(self, url='/status.xml', debug=True):
        """
        getRelay sets and reads data from the given host ip using the given
        user and password

        :param url: web address of relay box
        :param debug: write extended debug output
        :return: result: return values from web interface of box
        """

        if self.host is None:
            return None

        if not self.mutexPoll.tryLock():
            return None

        auth = requests.auth.HTTPBasicAuth(self._user,
                                           self._password,
                                           )
        url = f'http://{self._host[0]}:{self._host[1]}{url}'
        result = None

        try:
            result = requests.get(url, auth=auth, timeout=self.TIMEOUT)
        except requests.exceptions.Timeout:
            pass
        except requests.exceptions.ConnectionError:
            pass
        except Exception as e:
            self.logger.error(f'Error in request: {e}')

        if debug:
            self.debugOutput(result=result)

        self.mutexPoll.unlock()
        return result

    def cyclePolling(self):
        """
        cyclePolling reads the status of the relay status of each single relay.
        with success the statusReady single is sent.

        :return: success
        """

        value = self.getRelay('/status.xml', debug=False)

        if value is None:
            self.logger.error('Polling error')
            return False
        if value.reason != 'OK':
            self.logger.error('Polling error')
            return False

        lines = value.text.splitlines()
        for line in lines:
            value = re.findall(r'\d', line)
            if not value:
                continue
            value = [int(s) for s in value]
            self.status[value[0] - 1] = value[1]

        self.statusReady.emit()
        return True

    def getByte(self, relayNumber=0, state=False):
        """
        getByte generates the right bit mask for setting or resetting the relay mask

        :param relayNumber: relay number
        :param state: state to archive
        :return: bit mask for switching
        """

        byteStat = 0b0
        for i, status in enumerate(self.status):
            if status:
                byteStat = byteStat | 1 << i
        position = 1 << relayNumber
        byteOn = byteStat | position
        byteOff = byteOn & ~position

        if state:
            return byteOn
        else:
            return byteOff

    def pulse(self, relayNumber):
        """
        pulse switches a relay on for one second and off back.

        :param relayNumber: number of relay to be pulsed, counting from 0 onwards
        :return: success
        """

        self.logger.info(f'Pulse relay:{relayNumber}')
        byteOn = self.getByte(relayNumber=relayNumber, state=True)
        byteOff = self.getByte(relayNumber=relayNumber, state=False)
        value1 = self.getRelay(f'/FFE0{byteOn:02X}')
        time.sleep(self.PULSEWIDTH)
        value2 = self.getRelay(f'/FFE0{byteOff:02X}')

        if value1 is None or value2 is None:
            self.logger.error(f'Relay:{relayNumber}')
            return False
        elif value1.reason != 'OK' or value2.reason != 'OK':
            self.logger.error(f'Relay:{relayNumber}')
            return False

        return True

    def switch(self, relayNumber):
        """
        switch toggles the relay status (on, off)

        :param relayNumber: number of relay to be pulsed, counting from 0 onwards
        :return: success
        """

        self.logger.info(f'Switch relay:{relayNumber}')
        value = self.getRelay('/relays.cgi?relay={0:1d}'.format(relayNumber + 1))

        if value is None:
            self.logger.error(f'Relay:{relayNumber}')
            return False
        elif value.reason != 'OK':
            self.logger.error(f'Relay:{relayNumber}')
            return False

        return True

    def set(self, relayNumber, value):
        """
        set toggles the relay status to the desired value (on, off)

        :param relayNumber: number of relay to be pulsed, counting from 0 onwards
        :param value: relay state.
        :return: success
        """

        self.logger.info(f'Set relay:{relayNumber}')
        byteOn = self.getByte(relayNumber=relayNumber, state=value)
        value = self.getRelay(f'/FFE0{byteOn:02X}')

        if value is None:
            self.logger.error(f'Relay:{relayNumber}')
            return False
        elif value.reason != 'OK':
            self.logger.error(f'Relay:{relayNumber}')
            return False

        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
# Python  v3.7.4
#
# Michael Würtenberger
# (c) 2019
#
# Licence APL2.0
#
###########################################################
# standard libraries
# external packages
import PyQt5.QtCore
import PyQt5.QtWidgets
import PyQt5.uic
# local import


class SettDevice(object):
    """
    the main window class handles the main menu as well as the show and no show part of
    any other window. all necessary processing for functions of that gui will be linked
    to this class. therefore window classes will have a threadpool for managing async
    processing if needed.
    """

    def __init__(self):
        self.deviceDropDowns = [self.ui.imagingDevice,
                                self.ui.astrometryDevice,
                                self.ui.domeDevice,
                                self.ui.environDevice,
                                self.ui.skymeterDevice,
                                self.ui.powerDevice,
                                self.ui.relayDevice,
                                self.ui.measureDevice,
                                self.ui.remoteDevice,
                                ]
        self.deviceDropDownKeys = ['imagingDevice',
                                   'astrometryDevice',
                                   'domeDevice',
                                   'environmentDevice',
                                   'skymeterDevice',
                                   'powerDevice',
                                   'relayDevice',
                                   'measureDevice',
                                   'remoteDevice',
                                   ]
        self.setupDeviceGui()
        self.ui.imagingDevice.activated.connect(self.imagingDispatch)
        self.ui.relayDevice.activated.connect(self.relayDispatch)
        self.ui.remoteDevice.activated.connect(self.remoteDispatch)
        self.ui.measureDevice.activated.connect(self.measureDispatch)
        self.ui.domeDevice.activated.connect(self.domeDispatch)
        self.ui.environDevice.activated.connect(self.environDispatch)
        self.ui.skymeterDevice.activated.connect(self.skymeterDispatch)
        self.ui.powerDevice.activated.connect(self.powerDispatch)
        self.ui.astrometryDevice.activated.connect(self.astrometryDispatch)

    def initConfig(self):
        """
        initConfig read the key out of the configuration dict and stores it to the gui
        elements. if some initialisations have to be proceeded with the loaded persistent
        data, they will be launched as well in this method.

        :return: True for test purpose
        """
        config = self.app.config['mainW']
        for dropDown, key in zip(self.deviceDropDowns, self.deviceDropDownKeys):
            dropDown.setCurrentIndex(config.get(key, 0))

        self.relayDispatch()
        self.remoteDispatch()
        self.measureDispatch()
        self.domeDispatch()
        self.imagingDispatch()
        self.environDispatch()
        self.skymeterDispatch()
        self.powerDispatch()
        self.astrometryDispatch()
        return True

    def storeConfig(self):
        """
        storeConfig writes the keys to the configuration dict and stores. if some
        saving has to be proceeded to persistent data, they will be launched as
        well in this method.

        :return: True for test purpose
        """
        config = self.app.config['mainW']
        for dropDown, key in zip(self.deviceDropDowns, self.deviceDropDownKeys):
            config[key] = dropDown.currentIndex()

        return True

    def setupIcons(self):
        """
        setupIcons add icon from standard library to certain buttons for improving the
        gui of the app.

        :return:    True if success for test
        """
        return True

    def setupDeviceGui(self):
        """
        setupRelayGui handles the dropdown lists for all devices possible in mountwizzard.
        therefore we add the necessary entries to populate the list.

        :return: success for test
        """

        for dropDown in self.deviceDropDowns:
            dropDown.clear()
            dropDown.setView(PyQt5.QtWidgets.QListView())
            dropDown.addItem('No device selected')

        # adding special items
        self.ui.measureDevice.addItem('Built-In')
        self.ui.remoteDevice.addItem('Built-In')
        self.ui.relayDevice.addItem('Built-In')
        self.ui.environDevice.addItem('INDI')
        self.ui.domeDevice.addItem('INDI')
        self.ui.imagingDevice.addItem('INDI')
        self.ui.skymeterDevice.addItem('INDI')
        self.ui.powerDevice.addItem('INDI')
        for app in self.app.astrometry.available:
            self.ui.astrometryDevice.addItem(app)

        return True

    def relayDispatch(self):
        """
        relayDispatch allows to run the relay box.

        :return: success for test
        """

        # get index for relay tab
        tabWidget = self.ui.mainTabWidget.findChild(PyQt5.QtWidgets.QWidget, 'Relay')
        tabIndex = self.ui.mainTabWidget.indexOf(tabWidget)
        tabWidget2 = self.ui.settingsTabWidget.findChild(PyQt5.QtWidgets.QWidget, 'KMTronic')
        tabIndex2 = self.ui.settingsTabWidget.indexOf(tabWidget2)

        if self.ui.relayDevice.currentText().startswith('Built-In'):
            self.ui.mainTabWidget.setTabEnabled(tabIndex, True)
            self.ui.settingsTabWidget.setTabEnabled(tabIndex2, True)
            self.ui.mainTabWidget.setStyleSheet(self.getStyle())
            self.ui.settingsTabWidget.setStyleSheet(self.getStyle())
            self.app.message.emit('Relay enabled', 0)
            self.app.relay.startTimers()
            self.ui.relayDevice.setStyleSheet(self.BACK_GREEN)
        else:
            self.ui.mainTabWidget.setTabEnabled(tabIndex, False)
            self.ui.settingsTabWidget.setTabEnabled(tabIndex2, False)
            self.ui.mainTabWidget.setStyleSheet(self.getStyle())
            self.ui.settingsTabWidget.setStyleSheet(self.getStyle())
            self.app.message.emit('Relay disabled', 0)
            self.app.relay.stopTimers()
            self.ui.relayDevice.setStyleSheet(self.BACK_NORM)

        # update the style for showing the Relay tab
        self.ui.mainTabWidget.style().unpolish(self.ui.mainTabWidget)
        self.ui.mainTabWidget.style().polish(self.ui.mainTabWidget)
        self.ui.settingsTabWidget.style().unpolish(self.ui.settingsTabWidget)
        self.ui.settingsTabWidget.style().polish(self.ui.settingsTabWidget)
        return True

    def remoteDispatch(self):
        """
        remoteAccess enables or disables the remote access

        :return: true for test purpose
        """

        if self.ui.remoteDevice.currentText() == 'Built-In':
            self.app.remote.startRemote()
            self.app.message.emit('Remote enabled', 0)
            self.ui.remoteDevice.setStyleSheet(self.BACK_GREEN)
        else:
            self.app.remote.stopRemote()
            self.app.message.emit('Remote disabled', 0)
            self.ui.remoteDevice.setStyleSheet(self.BACK_NORM)

        return True

    def measureDispatch(self):
        """
        measureDispatch enables or disables the on board measurement process

        :return: true for test purpose
        """

        if self.ui.measureDevice.currentText().startswith('Built-In'):
            self.app.measure.startMeasurement()
            self.app.message.emit('Measurement enabled', 0)
            self.ui.measureDevice.setStyleSheet(self.BACK_GREEN)
        else:
            self.app.measure.stopMeasurement()
            self.app.message.emit('Measurement disabled', 0)
            self.ui.measureDevice.setStyleSheet(self.BACK_NORM)

        return True

    def domeDispatch(self):
        """
        domeDispatch selects the type of device for dome measures and start / stop
        them.
        in addition this function enables and disables other gui functions, which rely on
        the presence of a running driver

        :return: true for test purpose
        """

        self.app.dome.stopCommunication()
        if self.ui.domeDevice.currentText().startswith('INDI'):
            self.app.dome.client.host = self.ui.domeHost.text()
            self.app.dome.name = self.ui.domeDeviceName.currentText()
            self.app.dome.startCommunication()
            self.app.message.emit('Dome enabled', 0)
            self.deviceStat['dome'] = True
        else:
            self.app.message.emit('Dome disabled', 0)
            self.deviceStat['dome'] = None

        return True

    def imagingDispatch(self):
        """
        imagingDispatch selects the type of device for imaging and start / stop them.
        in addition this function enables and disables other gui functions, which rely on
        the presence of a running driver

        :return: true for test purpose
        """

        self.app.imaging.stopCommunication()
        if self.ui.imagingDevice.currentText().startswith('INDI'):
            self.app.imaging.client.host = self.ui.imagingHost.text()
            self.app.imaging.name = self.ui.imagingDeviceName.currentText()
            self.app.imaging.startCommunication()
            self.deviceStat['imaging'] = True
            self.app.message.emit('Imaging enabled', 0)
        else:
            self.deviceStat['imaging'] = None
            self.app.message.emit('Imaging disabled', 0)

        return True

    def environDispatch(self):
        """
        environDispatch selects the type of device for environment measures and start / stop
        them.
        in addition this function enables and disables other gui functions, which rely on
        the presence of a running driver

        :return: true for test purpose
        """

        self.app.environ.stopCommunication()
        if self.ui.environDevice.currentText().startswith('INDI'):
            self.app.environ.client.host = self.ui.environHost.text()
            self.app.environ.name = self.ui.environDeviceName.currentText()
            self.app.environ.startCommunication()
            self.deviceStat['environment'] = True
            self.app.message.emit('Environment enabled', 0)
        else:
            self.deviceStat['environment'] = None
            self.app.message.emit('Environment disabled', 0)

        return True

    def skymeterDispatch(self):
        """
        skymeterDispatch selects the type of device for environment measures and start / stop
        them.

        :return: true for test purpose
        """

        self.app.skymeter.stopCommunication()
        if self.ui.skymeterDevice.currentText().startswith('INDI'):
            self.app.skymeter.client.host = self.ui.skymeterHost.text()
            self.app.skymeter.name = self.ui.skymeterDeviceName.currentText()
            self.app.skymeter.startCommunication()
            self.app.message.emit('Skymeter enabled', 0)
            self.deviceStat['skymeter'] = True
        else:
            self.app.message.emit('Skymeter disabled', 0)
            self.deviceStat['skymeter'] = None

        return True

    def powerDispatch(self):
        """
        powerDispatch selects the type of device for environment measures and start / stop
        them.

        :return: true for test purpose
        """

        # get index for power tab
        tabWidget = self.ui.mainTabWidget.findChild(PyQt5.QtWidgets.QWidget, 'Power')
        tabIndex = self.ui.mainTabWidget.indexOf(tabWidget)

        self.app.power.stopCommunication()
        if self.ui.powerDevice.currentText().startswith('INDI'):
            self.app.power.startCommunication()
            self.ui.mainTabWidget.setTabEnabled(tabIndex, True)
            self.ui.mainTabWidget.setStyleSheet(self.getStyle())
            self.app.message.emit('Power enabled', 0)
            self.app.power.client.host = self.ui.powerHost.text()
            self.app.power.name = self.ui.powerDeviceName.currentText()
            self.deviceStat['power'] = True
        else:
            self.ui.mainTabWidget.setTabEnabled(tabIndex, False)
            self.ui.mainTabWidget.setStyleSheet(self.getStyle())
            self.app.message.emit('Power disabled', 0)
            self.deviceStat['power'] = None

        # update the style for showing the Relay tab
        self.ui.mainTabWidget.style().unpolish(self.ui.mainTabWidget)
        self.ui.mainTabWidget.style().polish(self.ui.mainTabWidget)
        return True

    def astrometryDispatch(self):
        """
        astrometryDispatch selects the type of device for environment measures and start / stop
        them.

        :return: true for test purpose
        """

        if self.ui.astrometryDevice.currentText() in self.app.astrometry.available:
            if self.ui.astrometryDevice.currentText() not in self.app.astrometry.binPath:
                self.app.message.emit('Astrometry not available', 2)
                self.deviceStat['astrometry'] = False
                return False
            self.ui.astrometryDevice.setStyleSheet(self.BACK_GREEN)
            self.deviceStat['astrometry'] = True
            self.app.message.emit('Astrometry enabled', 0)
        else:
            self.app.message.emit('Astrometry disabled', 0)
            self.deviceStat['astrometry'] = None
            self.ui.astrometryDevice.setStyleSheet(self.BACK_NORM)

        return True



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '../mw4/gui/widgets/message.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MessageDialog(object):
    def setupUi(self, MessageDialog):
        MessageDialog.setObjectName("MessageDialog")
        MessageDialog.resize(800, 600)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(MessageDialog.sizePolicy().hasHeightForWidth())
        MessageDialog.setSizePolicy(sizePolicy)
        MessageDialog.setMinimumSize(QtCore.QSize(800, 200))
        MessageDialog.setMaximumSize(QtCore.QSize(800, 1200))
        MessageDialog.setSizeIncrement(QtCore.QSize(10, 10))
        MessageDialog.setBaseSize(QtCore.QSize(10, 10))
        font = QtGui.QFont()
        font.setFamily("Arial")
        MessageDialog.setFont(font)
        self.verticalLayout = QtWidgets.QVBoxLayout(MessageDialog)
        self.verticalLayout.setContentsMargins(4, 4, 4, 4)
        self.verticalLayout.setSpacing(4)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setContentsMargins(4, 4, 4, 4)
        self.horizontalLayout.setSpacing(4)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.clear = QtWidgets.QPushButton(MessageDialog)
        self.clear.setMinimumSize(QtCore.QSize(80, 25))
        self.clear.setMaximumSize(QtCore.QSize(100, 25))
        self.clear.setObjectName("clear")
        self.horizontalLayout.addWidget(self.clear)
        self.widget = QtWidgets.QWidget(MessageDialog)
        self.widget.setObjectName("widget")
        self.horizontalLayout.addWidget(self.widget)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.line = QtWidgets.QFrame(MessageDialog)
        self.line.setFrameShadow(QtWidgets.QFrame.Plain)
        self.line.setLineWidth(2)
        self.line.setMidLineWidth(1)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)
        self.message = QtWidgets.QTextBrowser(MessageDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.message.sizePolicy().hasHeightForWidth())
        self.message.setSizePolicy(sizePolicy)
        self.message.setMinimumSize(QtCore.QSize(790, 190))
        self.message.setMaximumSize(QtCore.QSize(790, 1190))
        self.message.setSizeIncrement(QtCore.QSize(10, 10))
        self.message.setBaseSize(QtCore.QSize(10, 10))
        font = QtGui.QFont()
        font.setFamily("Courier New")
        font.setPointSize(10)
        font.setBold(False)
        font.setWeight(50)
        self.message.setFont(font)
        self.message.setAcceptDrops(False)
        self.message.setLocale(QtCore.QLocale(QtCore.QLocale.English, QtCore.QLocale.UnitedStates))
        self.message.setFrameShadow(QtWidgets.QFrame.Plain)
        self.message.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
        self.message.setHtml("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Courier New\'; font-size:10pt; font-weight:400; font-style:normal;\">\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:\'Arial\'; font-weight:600;\"><br /></p></body></html>")
        self.message.setAcceptRichText(False)
        self.message.setTextInteractionFlags(QtCore.Qt.NoTextInteraction)
        self.message.setPlaceholderText("")
        self.message.setSearchPaths([])
        self.message.setOpenLinks(False)
        self.message.setObjectName("message")
        self.verticalLayout.addWidget(self.message)
        self.verticalLayout.setStretch(2, 1)

        self.retranslateUi(MessageDialog)
        QtCore.QMetaObject.connectSlotsByName(MessageDialog)

    def retranslateUi(self, MessageDialog):
        _translate = QtCore.QCoreApplication.translate
        MessageDialog.setWindowTitle(_translate("MessageDialog", "Messages"))
        self.clear.setText(_translate("MessageDialog", "Clear window"))
        self.message.setToolTip(_translate("MessageDialog", "Error Messages from Tool"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    MessageDialog = QtWidgets.QWidget()
    ui = Ui_MessageDialog()
    ui.setupUi(MessageDialog)
    MessageDialog.show()
    sys.exit(app.exec_())



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
# Python  v3.7.4
#
# Michael Würtenberger
# (c) 2019
#
# Licence APL2.0
#
###########################################################
# standard libraries
# external packages
# local import


class Relay(object):
    """
    the main window class handles the main menu as well as the show and no show part of
    any other window. all necessary processing for functions of that gui will be linked
    to this class. therefore window classes will have a threadpool for managing async
    processing if needed.
    """

    def __init__(self):
        self.app.relay.statusReady.connect(self.updateRelayGui)

    def initConfig(self):
        # config = self.app.config['mainW']
        return True

    def storeConfig(self):
        # config = self.app.config['mainW']
        return True

    def setupIcons(self):
        """
        setupIcons add icon from standard library to certain buttons for improving the
        gui of the app.

        :return:    True if success for test
        """
        return True

    def updateRelayGui(self):
        """
        updateRelayGui changes the style of the button related to the state of the relay

        :return: success for test
        """

        for status, button in zip(self.app.relay.status, self.relayButtons):
            if status:
                self.changeStyleDynamic(button, 'running', 'true')
            else:
                self.changeStyleDynamic(button, 'running', 'false')
        return True



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '../mw4/gui/widgets/hemisphere.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_HemisphereDialog(object):
    def setupUi(self, HemisphereDialog):
        HemisphereDialog.setObjectName("HemisphereDialog")
        HemisphereDialog.resize(800, 600)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(HemisphereDialog.sizePolicy().hasHeightForWidth())
        HemisphereDialog.setSizePolicy(sizePolicy)
        HemisphereDialog.setMinimumSize(QtCore.QSize(800, 400))
        HemisphereDialog.setMaximumSize(QtCore.QSize(1600, 1200))
        HemisphereDialog.setSizeIncrement(QtCore.QSize(10, 10))
        HemisphereDialog.setBaseSize(QtCore.QSize(10, 10))
        font = QtGui.QFont()
        font.setFamily("Arial")
        HemisphereDialog.setFont(font)
        self.verticalLayout = QtWidgets.QVBoxLayout(HemisphereDialog)
        self.verticalLayout.setContentsMargins(4, 4, 4, 4)
        self.verticalLayout.setSpacing(4)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setContentsMargins(4, 4, 4, 4)
        self.horizontalLayout.setSpacing(4)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.groupBox_3 = QtWidgets.QGroupBox(HemisphereDialog)
        self.groupBox_3.setObjectName("groupBox_3")
        self.gridLayout = QtWidgets.QGridLayout(self.groupBox_3)
        self.gridLayout.setContentsMargins(10, 10, 10, 10)
        self.gridLayout.setSpacing(10)
        self.gridLayout.setObjectName("gridLayout")
        self.checkShowMeridian = QtWidgets.QCheckBox(self.groupBox_3)
        font = QtGui.QFont()
        font.setPointSize(10)
        self.checkShowMeridian.setFont(font)
        self.checkShowMeridian.setObjectName("checkShowMeridian")
        self.gridLayout.addWidget(self.checkShowMeridian, 1, 0, 1, 1)
        self.checkShowCelestial = QtWidgets.QCheckBox(self.groupBox_3)
        font = QtGui.QFont()
        font.setPointSize(10)
        self.checkShowCelestial.setFont(font)
        self.checkShowCelestial.setObjectName("checkShowCelestial")
        self.gridLayout.addWidget(self.checkShowCelestial, 2, 0, 1, 1)
        self.checkUseHorizon = QtWidgets.QCheckBox(self.groupBox_3)
        font = QtGui.QFont()
        font.setPointSize(10)
        self.checkUseHorizon.setFont(font)
        self.checkUseHorizon.setChecked(False)
        self.checkUseHorizon.setObjectName("checkUseHorizon")
        self.gridLayout.addWidget(self.checkUseHorizon, 0, 1, 1, 1)
        self.checkShowAlignStar = QtWidgets.QCheckBox(self.groupBox_3)
        font = QtGui.QFont()
        font.setPointSize(10)
        self.checkShowAlignStar.setFont(font)
        self.checkShowAlignStar.setObjectName("checkShowAlignStar")
        self.gridLayout.addWidget(self.checkShowAlignStar, 0, 0, 1, 1)
        self.checkShowSlewPath = QtWidgets.QCheckBox(self.groupBox_3)
        font = QtGui.QFont()
        font.setPointSize(10)
        self.checkShowSlewPath.setFont(font)
        self.checkShowSlewPath.setObjectName("checkShowSlewPath")
        self.gridLayout.addWidget(self.checkShowSlewPath, 3, 0, 1, 1)
        self.horizontalLayout.addWidget(self.groupBox_3)
        self.line_3 = QtWidgets.QFrame(HemisphereDialog)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Plain)
        self.line_3.setMidLineWidth(1)
        self.line_3.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_3.setObjectName("line_3")
        self.horizontalLayout.addWidget(self.line_3)
        self.groupBox_2 = QtWidgets.QGroupBox(HemisphereDialog)
        self.groupBox_2.setObjectName("groupBox_2")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.groupBox_2)
        self.gridLayout_2.setContentsMargins(10, 10, 10, 10)
        self.gridLayout_2.setSpacing(10)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.checkEditNone = QtWidgets.QRadioButton(self.groupBox_2)
        font = QtGui.QFont()
        font.setPointSize(10)
        self.checkEditNone.setFont(font)
        self.checkEditNone.setChecked(True)
        self.checkEditNone.setObjectName("checkEditNone")
        self.gridLayout_2.addWidget(self.checkEditNone, 0, 0, 1, 1)
        self.checkEditBuildPoints = QtWidgets.QRadioButton(self.groupBox_2)
        font = QtGui.QFont()
        font.setPointSize(10)
        self.checkEditBuildPoints.setFont(font)
        self.checkEditBuildPoints.setObjectName("checkEditBuildPoints")
        self.gridLayout_2.addWidget(self.checkEditBuildPoints, 1, 0, 1, 1)
        self.checkEditHorizonMask = QtWidgets.QRadioButton(self.groupBox_2)
        font = QtGui.QFont()
        font.setPointSize(10)
        self.checkEditHorizonMask.setFont(font)
        self.checkEditHorizonMask.setObjectName("checkEditHorizonMask")
        self.gridLayout_2.addWidget(self.checkEditHorizonMask, 2, 0, 1, 1)
        self.checkPolarAlignment = QtWidgets.QRadioButton(self.groupBox_2)
        self.checkPolarAlignment.setEnabled(True)
        font = QtGui.QFont()
        font.setPointSize(10)
        self.checkPolarAlignment.setFont(font)
        self.checkPolarAlignment.setObjectName("checkPolarAlignment")
        self.gridLayout_2.addWidget(self.checkPolarAlignment, 3, 0, 1, 1)
        self.horizontalLayout.addWidget(self.groupBox_2)
        self.line_4 = QtWidgets.QFrame(HemisphereDialog)
        self.line_4.setFrameShadow(QtWidgets.QFrame.Plain)
        self.line_4.setMidLineWidth(1)
        self.line_4.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_4.setObjectName("line_4")
        self.horizontalLayout.addWidget(self.line_4)
        self.groupBox = QtWidgets.QGroupBox(HemisphereDialog)
        self.groupBox.setEnabled(True)
        self.groupBox.setObjectName("groupBox")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.groupBox)
        self.gridLayout_4.setContentsMargins(10, 10, 10, 10)
        self.gridLayout_4.setSpacing(10)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.showAltAz = QtWidgets.QRadioButton(self.groupBox)
        self.showAltAz.setEnabled(True)
        font = QtGui.QFont()
        font.setPointSize(10)
        self.showAltAz.setFont(font)
        self.showAltAz.setObjectName("showAltAz")
        self.gridLayout_4.addWidget(self.showAltAz, 0, 0, 1, 1)
        self.showPolar = QtWidgets.QRadioButton(self.groupBox)
        self.showPolar.setEnabled(False)
        font = QtGui.QFont()
        font.setPointSize(10)
        self.showPolar.setFont(font)
        self.showPolar.setObjectName("showPolar")
        self.gridLayout_4.addWidget(self.showPolar, 1, 0, 1, 1)
        self.showSphere = QtWidgets.QRadioButton(self.groupBox)
        self.showSphere.setEnabled(False)
        font = QtGui.QFont()
        font.setPointSize(10)
        self.showSphere.setFont(font)
        self.showSphere.setObjectName("showSphere")
        self.gridLayout_4.addWidget(self.showSphere, 2, 0, 1, 1)
        self.horizontalLayout.addWidget(self.groupBox)
        self.line_5 = QtWidgets.QFrame(HemisphereDialog)
        self.line_5.setFrameShadow(QtWidgets.QFrame.Plain)
        self.line_5.setMidLineWidth(1)
        self.line_5.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_5.setObjectName("line_5")
        self.horizontalLayout.addWidget(self.line_5)
        self.gridLayout_3 = QtWidgets.QGridLayout()
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.clearBuildP = QtWidgets.QPushButton(HemisphereDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.clearBuildP.sizePolicy().hasHeightForWidth())
        self.clearBuildP.setSizePolicy(sizePolicy)
        self.clearBuildP.setMinimumSize(QtCore.QSize(100, 25))
        font = QtGui.QFont()
        font.setPointSize(10)
        self.clearBuildP.setFont(font)
        self.clearBuildP.setObjectName("clearBuildP")
        self.gridLayout_3.addWidget(self.clearBuildP, 0, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_3.addItem(spacerItem, 1, 0, 1, 1)
        self.horizontalLayout.addLayout(self.gridLayout_3)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.line = QtWidgets.QFrame(HemisphereDialog)
        self.line.setFrameShadow(QtWidgets.QFrame.Plain)
        self.line.setLineWidth(2)
        self.line.setMidLineWidth(1)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)
        self.hemisphere = QtWidgets.QWidget(HemisphereDialog)
        self.hemisphere.setEnabled(True)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.hemisphere.sizePolicy().hasHeightForWidth())
        self.hemisphere.setSizePolicy(sizePolicy)
        self.hemisphere.setMinimumSize(QtCore.QSize(790, 270))
        self.hemisphere.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.hemisphere.setSizeIncrement(QtCore.QSize(10, 10))
        self.hemisphere.setBaseSize(QtCore.QSize(10, 10))
        self.hemisphere.setAutoFillBackground(True)
        self.hemisphere.setStyleSheet("")
        self.hemisphere.setObjectName("hemisphere")
        self.verticalLayout.addWidget(self.hemisphere)
        self.verticalLayout.setStretch(2, 1)

        self.retranslateUi(HemisphereDialog)
        QtCore.QMetaObject.connectSlotsByName(HemisphereDialog)

    def retranslateUi(self, HemisphereDialog):
        _translate = QtCore.QCoreApplication.translate
        HemisphereDialog.setWindowTitle(_translate("HemisphereDialog", "Hemisphere"))
        self.groupBox_3.setTitle(_translate("HemisphereDialog", "Show Elements"))
        self.checkShowMeridian.setText(_translate("HemisphereDialog", "Meridian limits "))
        self.checkShowCelestial.setText(_translate("HemisphereDialog", "Celestial paths "))
        self.checkUseHorizon.setToolTip(_translate("HemisphereDialog", "<html><head/><body><p>Use horizon mask file in hemisphere window.</p></body></html>"))
        self.checkUseHorizon.setText(_translate("HemisphereDialog", "Horizon mask "))
        self.checkShowAlignStar.setText(_translate("HemisphereDialog", "Align stars "))
        self.checkShowSlewPath.setText(_translate("HemisphereDialog", "Slew path "))
        self.groupBox_2.setTitle(_translate("HemisphereDialog", "Operation Mode"))
        self.checkEditNone.setText(_translate("HemisphereDialog", "Normal mode"))
        self.checkEditBuildPoints.setText(_translate("HemisphereDialog", "Edit Build Points"))
        self.checkEditHorizonMask.setText(_translate("HemisphereDialog", "Edit Horizon Mask "))
        self.checkPolarAlignment.setText(_translate("HemisphereDialog", "Polar/Ortho Align "))
        self.groupBox.setTitle(_translate("HemisphereDialog", "Views / Coords"))
        self.showAltAz.setText(_translate("HemisphereDialog", "Alt/Az Plane"))
        self.showPolar.setText(_translate("HemisphereDialog", "Polar Diagram "))
        self.showSphere.setText(_translate("HemisphereDialog", "Sphere"))
        self.clearBuildP.setToolTip(_translate("HemisphereDialog", "<html><head/><body><p>Delete all points on coordinate window including Base / Model</p></body></html>"))
        self.clearBuildP.setText(_translate("HemisphereDialog", "Clear Points"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    HemisphereDialog = QtWidgets.QWidget()
    ui = Ui_HemisphereDialog()
    ui.setupUi(HemisphereDialog)
    HemisphereDialog.show()
    sys.exit(app.exec_())



# coding: utf-8
"""Spatial transformation library.
Description
===========
Create and apply several spatial 2D and 3D transformations including similarity,
bilinear,  projective,  polynomial and affine transformation. You can determine
the over-,  well- and under-determined parameters with the least-squares method.
Create 2D and 3D rotation matrices.
Usage
=====
>>> tform = make_tform('similarity',  np.array([[1, 1],  [2, 2]]),
... np.array([[3, 4],  [10, 10]]))
>>> tform.params
array([-3.25,   3.25,  -2.75,  -3.25])
>>> tform.params_explicit
array([-3.25      ,  -2.75      ,   4.59619408,  -0.78539816])
>>> tform.fwd(np.array([[0,  0],  [100, 100]]))
array([[  -3.25,    -2.75],
       [ 646.75,    -2.75]])
>>> tform.inv(tform.fwd(np.array([[0,  0],  [100, 100]])))
array([[   0.,     0.],
       [ 100.,   100.]])
Reference
=========
"Nahbereichsphotogrammetrie - Grundlagen,  Methoden und Anwendungen",
    Thomas Luhmann,  2010
"""
import warnings
import numpy as np
import math


TRANSFORMATIONS = [
    'similarity',
    'bilinear',
    'projective',
    'polynomial',
    'affine',
]


def make_tform(ttype,  src,  dst):
    """
    Create spatial transformation.
    You can determine the over-,  well- and under-determined parameters
    with the least-squares method.
    The following transformation types are supported:
        NAME / TTYPE            DIM     NUM POINTS FOR EXACT SOLUTION
        similarity:              2D      2
        bilinear:               2D      4
        projective:             2D      4
        polynomial (order n):  2D      (n+1)*(n+2)/2
        affine:                 2D      3
        affine:                 3D      4
    Number of source must match number of destination coordinates.
    :param ttype: similarity,  bilinear,  projective,  polynomial,  affine
        transformation type
    :param src: :class:`numpy.array`
        Nx2 or Nx3 coordinate matrix of source coordinate system
    :param dst: :class:`numpy.array`
        Nx2 or Nx3 coordinate matrix of destination coordinate system
    :returns: :class:`Transformation`
    """

    ttype = ttype.lower()
    if ttype not in TRANSFORMATIONS:
        raise NotImplemented(
            'Your transformation type %s is not implemented' % ttype)
    params,  params_explicit = MFUNCS[ttype](src,  dst)
    return Transformation(ttype,  params,  params_explicit)


def make_similarity(src,  dst):
    """
    Determine parameters of 2D similarity transformation in the order:
        a0,  a1,  b0,  b1
    where the transformation is defined as:
        X = a0 + a1*x - b1*y
        Y = b0 + b1*x + a1*y
    You can determine the over-,  well- and under-determined parameters
    with the least-squares method.
    Explicit parameters are in the order:
        a0,  b0,  m,  alpha [radians]
    where the transformation is defined as:
        X = a0 + m*x*cos(alpha) - m*y*sin(alpha)
        Y = b0 + m*x*sin(alpha) + m*y*cos(alpha)
    :param src: :class:`numpy.array`
        Nx2 coordinate matrix of source coordinate system
    :param dst: :class:`numpy.array`
        Nx2 coordinate matrix of destination coordinate system
    :returns: params,  params_explicit
    """

    xs = src[:, 0]
    ys = src[:, 1]
    rows = src.shape[0]
    A = np.zeros((rows*2,  4))
    A[:rows, 0] = 1
    A[:rows, 1] = xs
    A[:rows, 3] = - ys
    A[rows:, 2] = 1
    A[rows:, 3] = xs
    A[rows:, 1] = ys
    b = np.zeros((rows*2, ))
    b[:rows] = dst[:, 0]
    b[rows:] = dst[:, 1]
    params = np.linalg.lstsq(A,  b)[0]
    #: determine explicit params
    a0,  b0 = params[0],  params[2]
    alpha = math.atan2(params[3],  params[1])
    m = params[1] / math.cos(alpha)
    params_explicit = np.array([a0,  b0,  m,  alpha])
    return params,  params_explicit


def similarity_transform(coords,  params,  inverse=False):
    """
    Apply similarity transformation.
    :param coords: :class:`numpy.array`
        Nx2 coordinate matrix of source coordinate system
    :param params: :class:`numpy.array`
        parameters returned by `make_tform`
    :param inverse: bool
        apply inverse transformation,  default is False
    :returns: :class:`numpy.array`
        transformed coordinates
    """

    a0,  a1,  b0,  b1 = params
    x = coords[:, 0]
    y = coords[:, 1]
    out = np.zeros(coords.shape)
    if inverse:
        out[:, 0] = (a1*(x-a0)+b1*(y-b0)) / (a1**2+b1**2)
        out[:, 1] = (a1*(y-b0)-b1*(x-a0)) / (a1**2+b1**2)
    else:
        out[:, 0] = a0+a1*x-b1*y
        out[:, 1] = b0+b1*x+a1*y
    return out


def make_bilinear(src,  dst):
    """
    Determine parameters of 2D bilinear transformation in the order:
        a0,  a1,  a2,  a3,  b0,  b1,  b2,  b3
    where the transformation is defined as:
        X = a0 + a1*x + a2*y + a3*x*y
        Y = b0 + b1*x + b2*y + b3*x*y
    You can determine the over-,  well- and under-determined parameters
    with the least-squares method.
    :param src: :class:`numpy.array`
        Nx2 coordinate matrix of source coordinate system
    :param dst: :class:`numpy.array`
        Nx2 coordinate matrix of destination coordinate system
    :returns: params,  None
    """

    xs = src[:, 0]
    ys = src[:, 1]
    # affine transformation is polynomial transformation of order 1
    rows = src.shape[0]
    A = np.zeros((rows*2,  8))
    A[:rows, 0] = 1
    A[:rows, 1] = xs
    A[:rows, 2] = ys
    A[:rows, 3] = xs*ys
    A[rows:, 4] = 1
    A[rows:, 5] = xs
    A[rows:, 6] = ys
    A[:rows, 7] = xs*ys
    b = np.zeros((rows*2, ))
    b[:rows] = dst[:, 0]
    b[rows:] = dst[:, 1]
    params = np.linalg.lstsq(A,  b)[0]
    return params,  None


def bilinear_transform(coords,  params,  inverse=False):
    """
    Apply bilinear transformation.
    :param coords: :class:`numpy.array`
        Nx2 coordinate matrix of source coordinate system
    :param params: :class:`numpy.array`
        parameters returned by `make_tform`
    :param inverse: bool
        apply inverse transformation,  default is False
    :returns: :class:`numpy.array`
        transformed coordinates
    """

    a0,  a1,  a2,  a3,  b0,  b1,  b2,  b3 = params
    x = coords[:, 0]
    y = coords[:, 1]
    out = np.zeros(coords.shape)
    if inverse:
        raise NotImplemented('There is no explicit way to do the inverse '
                             'transformation. Determine the inverse transformation parameters '
                             'and use the fwd transformation instead.')
    else:
        out[:, 0] = a0+a1*x+a2*y+a3*x*y
        out[:, 1] = b0+b1*x+b2*y+b3*x*y
    return out


def make_projective(src,  dst):
    """
    Determine parameters of 2D projective transformation in the order:
        a0,  a1,  a2,  b0,  b1,  b2,  c0,  c1
    where the transformation is defined as:
        X = (a0+a1*x+a2*y) / (1+c0*x+c1*y)
        Y = (b0+b1*x+b2*y) / (1+c0*x+c1*y)
    You can determine the over-,  well- and under-determined parameters
    with the least-squares method.
    :param src: :class:`numpy.array`
        Nx2 coordinate matrix of source coordinate system
    :param dst: :class:`numpy.array`
        Nx2 coordinate matrix of destination coordinate system
    :returns: params,  None
    """

    xs = src[:, 0]
    ys = src[:, 1]
    rows = src.shape[0]
    A = np.zeros((rows*2,  8))
    A[:rows, 0] = 1
    A[:rows, 1] = xs
    A[:rows, 2] = ys
    A[rows:, 3] = 1
    A[rows:, 4] = xs
    A[rows:, 5] = ys
    A[:rows, 6] = - dst[:, 0] * xs
    A[:rows, 7] = - dst[:, 0] * ys
    A[rows:, 6] = - dst[:, 1] * xs
    A[rows:, 7] = - dst[:, 1] * ys
    b = np.zeros((rows*2, ))
    b[:rows] = dst[:, 0]
    b[rows:] = dst[:, 1]
    params = np.linalg.lstsq(A,  b)[0]
    return params,  None


def projective_transform(coords,  params,  inverse=False):
    """
    Apply projective transformation.
    :param coords: :class:`numpy.array`
        Nx2 coordinate matrix of source coordinate system
    :param params: :class:`numpy.array`
        parameters returned by `make_tform`
    :param inverse: bool
        apply inverse transformation,  default is False
    :returns: :class:`numpy.array`
        transformed coordinates
    """

    a0,  a1,  a2,  b0,  b1,  b2,  c0,  c1 = params
    x = coords[:, 0]
    y = coords[:, 1]
    out = np.zeros(coords.shape)
    if inverse:
        out[:, 0] = (a2*b0-a0*b2+(b2-b0*c1)*x+(a0*c1-a2)*y) \
            / (a1*b2-a2*b1+(b1*c1-b2*c0)*x+(a2*c0-a1*c1)*y)
        out[:, 1] = (a0*b1-a1*b0+(b0*c0-b1)*x+(a1-a0*c0)*y) \
            / (a1*b2-a2*b1+(b1*c1-b2*c0)*x+(a2*c0-a1*c1)*y)
    else:
        out[:, 0] = (a0+a1*x+a2*y) / (1+c0*x+c1*y)
        out[:, 1] = (b0+b1*x+b2*y) / (1+c0*x+c1*y)
    return out


def make_polynomial(src,  dst,  n):
    """
    Determine parameters of 2D polynomial transformation of order n,
    where the transformation is defined as:
        X = sum[j=0:n](sum[i=0:j](a_ji * x**(j-i)*y**i))
        Y = sum[j=0:n](sum[i=0:j](b_ji * x**(j-i)*y**i))
    You can determine the over-,  well- and under-determined parameters
    with the least-squares method.
    :param src: :class:`numpy.array`
        Nx2 coordinate matrix of source coordinate system
    :param dst: :class:`numpy.array`
        Nx2 coordinate matrix of destination coordinate system
    :param n
    :returns: params,  None
    """

    xs = src[:, 0]
    ys = src[:, 1]
    # number of unknown coefficients
    u = (n+1)*(n+2)
    rows = src.shape[0]
    A = np.zeros((rows*2,  u))
    pidx = 0
    for j in xrange(n+1):
        for i in xrange(j+1):
            A[:rows, pidx] = xs**(j-i)*ys**i
            A[rows:, pidx+u/2] = xs**(j-i)*ys**i
            pidx += 1
    b = np.zeros((rows*2, ))
    b[:rows] = dst[:, 0]
    b[rows:] = dst[:, 1]
    params = np.linalg.lstsq(A,  b)[0]
    return params,  None


def polynomial_transform(coords,  params,  inverse=False):
    """
    Apply polynomial transformation.
    :param coords: :class:`numpy.array`
        Nx2 coordinate matrix of source coordinate system
    :param params: :class:`numpy.array`
        parameters returned by `make_tform`
    :param inverse: bool
        apply inverse transformation,  default is False
    :returns: :class:`numpy.array`
        transformed coordinates
    """

    x = coords[:, 0]
    y = coords[:, 1]
    u = len(params)
    # number of coefficients ( -> u = (n+1)*(n+2) )
    n = int((-3+math.sqrt(9-4*(2-u))) / 2)
    out = np.zeros(coords.shape)
    if inverse:
        raise NotImplemented('There is no explicit way to do the inverse '
                             'polynomial transformation as it is in general non-linear.'
                             'Determine the inverse transformation parameters '
                             'and use the fwd transformation instead.')
    else:
        pidx = 0
        for j in xrange(n+1):
            for i in xrange(j+1):
                out[:, 0] += params[pidx]*x**(j-i)*y**i
                out[:, 1] += params[pidx+u/2]*x**(j-i)*y**i
                pidx += 1
    return out


def make_affine(src,  dst):
    """
    Determine parameters of 2D or 3D affine transformation in the order:
        a0,  a1,  a2,  a3,  b0,  b1,  b2,  b3,  c0,  c1,  c2,  c3
    where the transformation is defined as:
        X = a0 + a1*x + a2*y[ + a3*z]
        Y = b0 + b1*x + b2*y[ + b3*z]
        [Z = c0 + c1*x + c2*y + c3*z]
    You can determine the over-,  well- and under-determined parameters
    with the least-squares method.
    Source and destination coordinates must be Nx2 or Nx3 matrices (x,  y,  z).
    Explicit parameters are in the order:
        a0,  b0,  c0,  mx,  my,  mz,  alpha [radians],  beta [radians],  gamma [radians]
    where the 3D transformation is defined as (excluding the :
        X = tx * R3(gamma)*R2(beta)*R1(alpha)*S*x
        with
            X = (X,  Y,  Z).T
            tx = (a0,  b0,  c0).T
            R1(alpha) = rotation_matrix(alpha,  axis=1)
            R2(beta) = rotation_matrix(beta,  axis=2)
            R3(gamma) = rotation_matrix(gamma,  axis=3)
            S = diag(mx,  my,  mz)
            x = (x,  y,  z).T
    and the simplified 2D transformation as:
        X = a0 + mx*x*cos(alpha) - my*y*sin(alpha+beta)
        Y = b0 + mx*x*sin(alpha) + my*y*cos(alpha+beta)
    In case of 2D coordinates the following parameters are 0:
        a3,  b3,  c0,  c1,  c2,  c3
    and the explicit parameters
        c0,  mz,  gamma
    :param src: :class:`numpy.array`
        Nx2 or Nx3 coordinate matrix of source coordinate system
    :param dst: :class:`numpy.array`
        Nx2 or Nx3 coordinate matrix of destination coordinate system
    :returns: params,  params_explicit
    """

    xs = src[:, 0]
    ys = src[:, 1]
    rows = src.shape[0]
    A = np.zeros((rows*3,  12))
    A[:rows, 0] = 1
    A[:rows, 1] = xs
    A[:rows, 2] = ys
    A[rows:rows*2, 4] = 1
    A[rows:rows*2, 5] = xs
    A[rows:rows*2, 6] = ys
    A[rows*2:, 8] = 1
    A[rows*2:, 9] = xs
    A[rows*2:, 10] = ys
    b = np.zeros((rows*3, ))
    b[:rows] = dst[:, 0]
    b[rows:rows*2] = dst[:, 1]
    if src.shape[1] == 3:
        zs = src[:, 2]
        A[:rows, 3] = zs
        A[rows:rows*2, 7] = zs
        A[rows*2:, 11] = zs
        b[rows*2:] = dst[:, 2]
    params = np.linalg.lstsq(A,  b)[0]
    #: determine explicit parameters
    a0,  a1,  a2,  a3,  b0,  b1,  b2,  b3,  c0,  c1,  c2,  c3 = params
    mx = math.sqrt(a1**2+b1**2+c1**2)
    my = math.sqrt(a2**2+b2**2+c2**2)
    mz = math.sqrt(a3**2+b3**2+c3**2)
    if src.shape[1] == 3 and 0 in (mx,  my,  mz):
        warnings.warn('One of your scale factors are 0,  you should probably '
                      'use a 2D instead of a 3D affine transformation.',
                      RuntimeWarning)
    alpha = math.atan2(c2,  c3)
    beta = math.atan2(-c1,  math.sqrt(a1**2+b1**2))
    gamma = math.atan2(b1,  a1)
    params_explicit = np.array([a0,  b0,  c0,  mx,  my,  mz,  alpha,  beta,  gamma])
    return params,  params_explicit


def affine_transform(coords,  params,  inverse=False):
    """
    Apply 2D or 3D affine transformation.
    :param coords: :class:`numpy.array`
        Nx2 coordinate matrix of source coordinate system
    :param params: :class:`numpy.array`
        parameters returned by `make_tform`
    :param inverse: bool
        apply inverse transformation,  default is False
    :returns: :class:`numpy.array`
        transformed coordinates
    """

    a0,  a1,  a2,  a3,  b0,  b1,  b2,  b3,  c0,  c1,  c2,  c3 = params
    x = coords[:, 0]
    y = coords[:, 1]
    out = np.zeros(coords.shape)
    if coords.shape[1] == 2:
        if inverse:
            out[:, 0] = (a2*(y-b0)-b2*(x-a0)) / (a2*b1-a1*b2)
            out[:, 1] = (b1*(x-a0)-a1*(y-b0)) / (a2*b1-a1*b2)
        else:
            out[:, 0] = a0+a1*x+a2*y
            out[:, 1] = b0+b1*x+b2*y
    elif coords.shape[1] == 3:
        z = coords[:, 2]
        if inverse:
            out[:, 0] = (
                    (b2*c3-b3*c2)*x - (b2*c3-b3*c2)*a0
                    + ((z-c0)*b3-y*c3+b0*c3)*a2 - ((z-c0)*b2-y*c2+b0*c2)*a3
                ) / ((b2*c3-b3*c2)*a1 - (b1*c3-b3*c1)*a2 + (b1*c2-b2*c1)*a3)
            out[:, 1] = -(
                    (b1*c3-b3*c1)*x - (b1*c3-b3*c1)*a0
                    + ((z-c0)*b3-y*c3+b0*c3)*a1 - ((z-c0)*b1-y*c1+b0*c1)*a3
                ) / ((b2*c3-b3*c2)*a1 - (b1*c3-b3*c1)*a2 + (b1*c2-b2*c1)*a3)
            out[:, 2] = (
                    (b1*c2-b2*c1)*x - (b1*c2-b2*c1)*a0
                    + ((z-c0)*b2-y*c2+b0*c2)*a1 - ((z-c0)*b1-y*c1+b0*c1)*a2
                ) / ((b2*c3-b3*c2)*a1 - (b1*c3-b3*c1)*a2 + (b1*c2-b2*c1)*a3)
        else:
            out[:, 0] = a0+a1*x+a2*y+a3*z
            out[:, 1] = b0+b1*x+b2*y+b3*z
            out[:, 2] = c0+c1*x+c2*y+c3*z
    return out


MFUNCS = {
    'similarity': make_similarity,
    'bilinear': make_bilinear,
    'projective': make_projective,
    'polynomial': make_polynomial,
    'affine': make_affine,
}
TFUNCS = {
    'similarity': similarity_transform,
    'bilinear': bilinear_transform,
    'projective': projective_transform,
    'polynomial': polynomial_transform,
    'affine': affine_transform,
}


class Transformation(object):

    def __init__(self,  ttype,  params,  params_explicit=None):
        """
        Create transformation which allows you to do forward and inverse
        transformation and view the transformation parameters.
        :param ttype: similarity,  bilinear,  projective,  polynomial,  affine
            transformation type
        :param params: :class:`numpy.array`
            transformation parameters
        :param params: :class:`numpy.array`
            explicit transformation parameters as
        """

        self.ttype = ttype
        self.params = params
        self.params_explicit = params_explicit

    def fwd(self,  coords):
        """
        Apply forward transformation.
        :param coords: :class:`numpy.array`
            Nx2 or Nx3 coordinate matrix
        """

        single = False
        if coords.ndim == 1:
            coords = np.array([coords])
            single = True
        result = TFUNCS[self.ttype](coords,  self.params,  inverse=False)
        if single:
            return result[0]
        return result

    def inv(self,  coords):
        """
        Apply inverse transformation.
        :param coords: :class:`numpy.array`
            Nx2 or Nx3 coordinate matrix
        """

        single = False
        if coords.ndim == 1:
            coords = np.array([coords])
            single = True
        result = TFUNCS[self.ttype](coords,  self.params,  inverse=True)
        if single:
            return result[0]
        return result


def rotation_matrix(angle,  dim=2,  axis=None):
    """
    Create a 2D or 3D rotation matrix.
    :param angle: or float as radians
        angle of rotation
    :param dim: 2 or 3,  optional
        dimension of rotation matrix,  default is 2
    :param axis: 1,  2 or 3,  optional
        rotation axis for 3D rotation,  default is None
    :returns: 2x2 or 3x3 rotation matrix
    """

    R = None
    if dim == 2:
        R = [
            [math.cos(angle),  -math.sin(angle)],
            [math.sin(angle),  math.cos(angle)],
        ]
    elif dim == 3:
        if axis == 1:
            R = [
                [1,  0,  0],
                [0,  math.cos(angle),  -math.sin(angle)],
                [0,  math.sin(angle),  math.cos(angle)],
            ]
        elif axis == 2:
            R = [
                [math.cos(angle),  0,  math.sin(angle)],
                [0,  1,  0],
                [-math.sin(angle),  0,  math.cos(angle)],
            ]
        elif axis == 3:
            R = [
                [math.cos(angle),  -math.sin(angle),  0],
                [math.sin(angle),  math.cos(angle),  0],
                [0,  0,  1],
            ]
    return np.array(R)



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
# Python  v3.7.4
#
# Michael Würtenberger
# (c) 2019
#
# Licence APL2.0
#
###########################################################
# standard libraries
import queue
import os
import time
import shutil
import json
from datetime import datetime, timedelta
# external packages
import PyQt5.QtWidgets
import PyQt5.uic
import skyfield.api
from mountcontrol.model import APoint
import numpy as np
# local import
from mw4.definitions import Point, MPoint, IParam, MParam, MData, RData, Solve, Solution
from mw4.base import transform


class QMultiWait(PyQt5.QtCore.QObject):
    """
    QMultiWaitable implements a signal collection class for waiting of entering multiple
    signals before firing the "AND" relation of all signals.
    derived from:

    https://stackoverflow.com/questions/21108407/qt-how-to-wait-for-multiple-signals

    in addition all received signals could be reset
    """

    ready = PyQt5.QtCore.pyqtSignal()

    def __init__(self):
        super().__init__()
        self.waitable = set()
        self.waitready = set()

    def addWaitableSignal(self, signal):
        if signal not in self.waitable:
            self.waitable.add(signal)
            signal.connect(self.checkSignal)

    def checkSignal(self):
        sender = self.sender()
        self.waitready.add(sender)
        if len(self.waitready) == len(self.waitable):
            self.ready.emit()

    def resetSignals(self):
        self.waitready = set()

    def clear(self):
        for signal in self.waitable:
            signal.disconnect(self.checkSignal)
        self.waitable = set()
        self.waitready = set()


class BuildModel(object):
    """
    the main window class handles the main menu as well as the show and no show part of
    any other window. all necessary processing for functions of that gui will be linked
    to this class. therefore window classes will have a threadpool for managing async
    processing if needed.
    """

    def __init__(self):

        self.slewQueue = queue.Queue()
        self.imageQueue = queue.Queue()
        self.solveQueue = queue.Queue()
        self.resultQueue = queue.Queue()
        self.modelQueue = queue.Queue()
        self.collector = QMultiWait()
        self.startModeling = None
        self.modelName = ''
        self.imageDir = ''
        self.lastModelType = ''

        self.ui.genBuildGrid.clicked.connect(self.genBuildGrid)
        self.ui.numberGridPointsCol.valueChanged.connect(self.genBuildGrid)
        self.ui.numberGridPointsRow.valueChanged.connect(self.genBuildGrid)
        self.ui.altitudeMin.valueChanged.connect(self.genBuildGrid)
        self.ui.altitudeMax.valueChanged.connect(self.genBuildGrid)
        self.ui.genBuildMax.clicked.connect(self.genBuildMax)
        self.ui.genBuildMed.clicked.connect(self.genBuildMed)
        self.ui.genBuildNorm.clicked.connect(self.genBuildNorm)
        self.ui.genBuildMin.clicked.connect(self.genBuildMin)
        self.ui.genBuildFile.clicked.connect(self.genBuildFile)
        self.ui.genBuildDSO.clicked.connect(self.genBuildDSO)
        self.ui.numberDSOPoints.valueChanged.connect(self.genBuildDSO)
        self.ui.durationDSO.valueChanged.connect(self.genBuildDSO)
        self.ui.timeShiftDSO.valueChanged.connect(self.genBuildDSO)
        self.ui.saveBuildPoints.clicked.connect(self.saveBuildFile)
        self.ui.saveBuildPointsAs.clicked.connect(self.saveBuildFileAs)
        self.ui.loadBuildPoints.clicked.connect(self.loadBuildFile)
        self.ui.numberSpiralPoints.valueChanged.connect(self.genBuildGoldenSpiral)
        self.ui.genBuildGoldenSpiral.clicked.connect(self.genBuildGoldenSpiral)
        self.ui.runFullModel.clicked.connect(self.modelFull)
        self.ui.cancelFullModel.clicked.connect(self.cancelFull)
        self.ui.runAlignModel.clicked.connect(self.modelAlign)
        self.ui.cancelAlignModel.clicked.connect(self.cancelFull)
        self.ui.batchModel.clicked.connect(self.loadProgramModel)

    def initConfig(self):
        """
        initConfig read the key out of the configuration dict and stores it to the gui
        elements. if some initialisations have to be proceeded with the loaded persistent
        data, they will be launched as well in this method.

        :return: True for test purpose
        """
        config = self.app.config['mainW']

        self.ui.numberGridPointsCol.valueChanged.disconnect(self.genBuildGrid)
        self.ui.numberGridPointsRow.valueChanged.disconnect(self.genBuildGrid)
        self.ui.altitudeMin.valueChanged.disconnect(self.genBuildGrid)
        self.ui.altitudeMax.valueChanged.disconnect(self.genBuildGrid)
        self.ui.numberDSOPoints.valueChanged.disconnect(self.genBuildDSO)
        self.ui.durationDSO.valueChanged.disconnect(self.genBuildDSO)
        self.ui.timeShiftDSO.valueChanged.disconnect(self.genBuildDSO)

        self.ui.buildPFileName.setText(config.get('buildPFileName', ''))
        self.lastModelType = config.get('lastModelType', '')
        self.ui.numberGridPointsRow.setValue(config.get('numberGridPointsRow', 5))
        self.ui.numberGridPointsCol.setValue(config.get('numberGridPointsCol', 6))
        self.ui.altitudeMin.setValue(config.get('altitudeMin', 30))
        self.ui.altitudeMax.setValue(config.get('altitudeMax', 75))
        self.ui.numberDSOPoints.setValue(config.get('numberDSOPoints', 20))
        self.ui.durationDSO.setValue(config.get('durationDSO', 5))
        self.ui.timeShiftDSO.setValue(config.get('timeShiftDSO', 0))

        # initialising the signal slot connections after the value are set, because
        # otherwise we get a first value changed signal just when populating
        # the initial data. this should not happen.
        self.ui.numberGridPointsCol.valueChanged.connect(self.genBuildGrid)
        self.ui.numberGridPointsRow.valueChanged.connect(self.genBuildGrid)
        self.ui.altitudeMin.valueChanged.connect(self.genBuildGrid)
        self.ui.altitudeMax.valueChanged.connect(self.genBuildGrid)
        self.ui.numberDSOPoints.valueChanged.connect(self.genBuildDSO)
        self.ui.durationDSO.valueChanged.connect(self.genBuildDSO)
        self.ui.timeShiftDSO.valueChanged.connect(self.genBuildDSO)

        return True

    def storeConfig(self):
        """
        storeConfig writes the keys to the configuration dict and stores. if some
        saving has to be proceeded to persistent data, they will be launched as
        well in this method.

        :return: True for test purpose
        """
        config = self.app.config['mainW']
        config['buildPFileName'] = self.ui.buildPFileName.text()
        config['lastModelType'] = self.lastModelType
        config['numberGridPointsRow'] = self.ui.numberGridPointsRow.value()
        config['numberGridPointsCol'] = self.ui.numberGridPointsCol.value()
        config['altitudeMin'] = self.ui.altitudeMin.value()
        config['altitudeMax'] = self.ui.altitudeMax.value()
        config['numberDSOPoints'] = self.ui.numberDSOPoints.value()
        config['durationDSO'] = self.ui.durationDSO.value()
        config['timeShiftDSO'] = self.ui.timeShiftDSO.value()

        return True

    def setupIcons(self):
        """
        setupIcons add icon from standard library to certain buttons for improving the
        gui of the app.

        :return:    True for test purpose
        """

        self.wIcon(self.ui.genBuildGrid, PyQt5.QtWidgets.QStyle.SP_DialogApplyButton)
        self.wIcon(self.ui.genBuildMax, PyQt5.QtWidgets.QStyle.SP_DialogApplyButton)
        self.wIcon(self.ui.genBuildMed, PyQt5.QtWidgets.QStyle.SP_DialogApplyButton)
        self.wIcon(self.ui.genBuildNorm, PyQt5.QtWidgets.QStyle.SP_DialogApplyButton)
        self.wIcon(self.ui.genBuildMin, PyQt5.QtWidgets.QStyle.SP_DialogApplyButton)
        self.wIcon(self.ui.genBuildDSO, PyQt5.QtWidgets.QStyle.SP_DialogApplyButton)

        return True

    def genBuildGrid(self):
        """
        genBuildGrid generates a grid of point for model build based on gui data. the cols
        have to be on even numbers.

        :return: success
        """

        self.lastModelType = 'grid'
        row = self.ui.numberGridPointsRow.value()
        col = self.ui.numberGridPointsCol.value()
        minAlt = self.ui.altitudeMin.value()
        maxAlt = self.ui.altitudeMax.value()
        suc = self.app.data.genGrid(minAlt=minAlt,
                                    maxAlt=maxAlt,
                                    numbRows=row,
                                    numbCols=col)
        if not suc:
            return False

        self.autoDeletePoints()
        self.autoSortPoints()

        return True

    def genBuildMax(self):
        """
        genBuildMax generates the point pattern based on greater circles for model build.
        the point are calculated for the observers position. max goes for approx 100 points
        effectively when removing the horizon.

        :return: success
        """

        self.lastModelType = 'circle-max'
        suc = self.app.data.genGreaterCircle(selection='max')
        if not suc:
            self.app.message.emit('Build points [max] cannot be generated', 2)
            return False

        self.autoDeletePoints()
        self.autoSortPoints()

        return True

    def genBuildMed(self):
        """
        genBuildMed generates the point pattern based on greater circles for model build.
        the point are calculated for the observers position. max goes for approx 70 points
        effectively when removing the horizon.

        :return: success
        """

        self.lastModelType = 'circle-med'
        suc = self.app.data.genGreaterCircle(selection='med')
        if not suc:
            self.app.message.emit('Build points [med] cannot be generated', 2)
            return False

        self.autoDeletePoints()
        self.autoSortPoints()

        return True

    def genBuildNorm(self):
        """
        genBuildNorm generates the point pattern based on greater circles for model build.
        the point are calculated for the observers position. max goes for approx 40 points
        effectively when removing the horizon.

        :return: success
        """

        self.lastModelType = 'circle-norm'
        suc = self.app.data.genGreaterCircle(selection='norm')
        if not suc:
            self.app.message.emit('Build points [norm] cannot be generated', 2)
            return False

        self.autoDeletePoints()
        self.autoSortPoints()

        return True

    def genBuildMin(self):
        """
        genBuildMin generates the point pattern based on greater circles for model build.
        the point are calculated for the observers position. min goes for approx 25 points
        effectively when removing the horizon.

        :return: success
        """

        self.lastModelType = 'circle-min'
        suc = self.app.data.genGreaterCircle(selection='min')
        if not suc:
            self.app.message.emit('Build points [min] cannot be generated', 2)
            return False

        self.autoDeletePoints()
        self.autoSortPoints()

        return True

    def genBuildDSO(self):
        """
        genBuildDSO generates points along the actual tracking path

        :return: success
        """

        self.lastModelType = 'DSO'
        ra = self.app.mount.obsSite.raJNow
        dec = self.app.mount.obsSite.decJNow
        timeJD = self.app.mount.obsSite.timeJD
        location = self.app.mount.obsSite.location

        if ra is None or dec is None or location is None:
            self.app.message.emit('DSO Path cannot be generated', 2)
            return False

        numberPoints = self.ui.numberDSOPoints.value()
        duration = self.ui.durationDSO.value()
        timeShift = self.ui.timeShiftDSO.value()

        suc = self.app.data.generateDSOPath(ra=ra,
                                            dec=dec,
                                            timeJD=timeJD,
                                            location=location,
                                            numberPoints=numberPoints,
                                            duration=duration,
                                            timeShift=timeShift,
                                            )

        if not suc:
            self.app.message.emit('DSO Path cannot be generated', 2)
            return False

        self.autoDeletePoints()
        self.autoSortPoints()

        return True

    def genBuildGoldenSpiral(self):
        """
        genBuildGoldenSpiral generates points along the actual tracking path

        :return: success
        """

        self.lastModelType = 'spiral'
        numberPoints = self.ui.numberSpiralPoints.value()

        suc = self.app.data.generateGoldenSpiral(numberPoints=numberPoints)
        if not suc:
            self.app.message.emit('Golden spiral cannot be generated', 2)
            return False

        self.autoDeletePoints()
        self.autoSortPoints()

        return True

    def loadBuildFile(self):
        """
        loadBuildFile calls a file selector box and selects the filename to be loaded

        :return: success
        """

        folder = self.app.mwGlob['configDir']
        loadFilePath, fileName, ext = self.openFile(self,
                                                    'Open build point file',
                                                    folder,
                                                    'Build point files (*.bpts)',
                                                    )
        if not loadFilePath:
            return False

        suc = self.app.data.loadBuildP(fileName=fileName)
        if suc:
            self.ui.buildPFileName.setText(fileName)
            self.app.message.emit('Build file [{0}] loaded'.format(fileName), 0)
        else:
            self.app.message.emit('Build file [{0}] cannot no be loaded'.format(fileName), 2)

        return True

    def saveBuildFile(self):
        """
        saveBuildFile calls saving the build file

        :return: success
        """

        fileName = self.ui.buildPFileName.text()
        if not fileName:
            self.app.message.emit('Build points file name not given', 2)
            return False

        suc = self.app.data.saveBuildP(fileName=fileName)

        if suc:
            self.app.message.emit('Build file [{0}] saved'.format(fileName), 0)
        else:
            self.app.message.emit('Build file [{0}] cannot no be saved'.format(fileName), 2)

        return True

    def saveBuildFileAs(self):
        """
        saveBuildFileAs calls a file selector box and selects the filename to be save

        :return: success
        """

        folder = self.app.mwGlob['configDir']
        saveFilePath, fileName, ext = self.saveFile(self,
                                                    'Save build point file',
                                                    folder,
                                                    'Build point files (*.bpts)',
                                                    )
        if not saveFilePath:
            return False

        suc = self.app.data.saveBuildP(fileName=fileName)

        if suc:
            self.ui.buildPFileName.setText(fileName)
            self.app.message.emit('Build file [{0}] saved'.format(fileName), 0)
        else:
            self.app.message.emit('Build file [{0}] cannot no be saved'.format(fileName), 2)

        return True

    def genBuildFile(self):
        """
        genBuildFile tries to load a give build point file and displays it for usage.

        :return: success
        """

        fileName = self.ui.buildPFileName.text()
        if not fileName:
            self.app.message.emit('Build points file name not given', 2)
            return False

        self.lastModelType = 'file'
        suc = self.app.data.loadBuildP(fileName=fileName)
        if not suc:
            text = 'Build points file [{0}] could not be loaded'.format(fileName)
            self.app.message.emit(text, 2)
            return False

        self.autoDeletePoints()

        return True

    def updateProgress(self, number=0, count=0, modelingDone=False):
        """
        updateProgress calculated from the elapsed time and number of point with taking
        actual processing time into account a estimation of duration and finishing time
        of the modeling process and updates this in the gui

        :param number: total number of model points
        :param count: index of the actual processed point
        :param modelingDone: state for the last point
        :return: success
        """

        if not 0 <= count < number:
            return False
        if not number:
            return False

        modelPercent = (count + 1) / number
        timeElapsed = time.time() - self.startModeling

        if modelingDone:
            timeEstimation = 0
        else:
            timeEstimation = (1 / modelPercent * timeElapsed) * (1 - modelPercent)
        finished = timedelta(seconds=timeEstimation) + datetime.now()

        self.ui.timeToFinish.setText(time.strftime('%M:%S', time.gmtime(timeEstimation)))
        self.ui.timeElapsed.setText(time.strftime('%M:%S', time.gmtime(timeElapsed)))
        self.ui.timeFinished.setText(finished.strftime('%H:%M:%S'))
        self.ui.modelProgress.setValue(modelPercent * 100)

        return True

    @staticmethod
    def addResultToModel(mPoint=None, result=None):
        """
        addResultToModel takes the result of the solving process and add the data to the
        actual model point. as the coordinates for programming the model need to be in
        JNow and the solving process gives J2000 coordinates, it does the transform as
        well.

        :param mPoint:
        :param result:
        :return: model
        """

        rData = result.solve
        raJNow, decJNow = transform.J2000ToJNow(rData.raJ2000,
                                                rData.decJ2000,
                                                mPoint.mData.julian)

        mData = MData(raMJNow=mPoint.mData.raMJNow,
                      decMJNow=mPoint.mData.decMJNow,
                      raSJNow=raJNow,
                      decSJNow=decJNow,
                      sidereal=mPoint.mData.sidereal,
                      julian=mPoint.mData.julian,
                      pierside=mPoint.mData.pierside,
                      )
        mPoint = MPoint(mParam=mPoint.mParam,
                        iParam=mPoint.iParam,
                        point=mPoint.point,
                        mData=mData,
                        rData=mPoint.rData,
                        )

        return mPoint

    def modelSolveDone(self, result):
        """
        modelSolveDone is called when a point is solved by astrometry. if called it takes
        the model point out of the queue and adds the solving data for later model build.
        as the solving takes place in J2000 epoch, but we need fpr die model build JNow
        epoch, the transformation is done as well.

        in addition as it is the last step before a model point could be used, the
        it checks for the end of the modeling process.

        :param result: true for test purpose
        :return: success
        """

        if self.resultQueue.empty():
            self.logger.error('empty result queue')
            return False

        mPoint = self.resultQueue.get()
        number = mPoint.mParam.number
        count = mPoint.mParam.count
        modelingDone = (number == count + 1)

        if not isinstance(result, Solution):
            self.logger.info(f'Solving result is malformed: {result}')
            self.logger.error('empty result queue')
            return False
        if not isinstance(result.solve, Solve):
            self.logger.info(f'Solving result is malformed: {result}')
            self.logger.error('empty result queue')
            return False

        # processing only the model point which are OK
        if result.success:
            mPoint = self.addResultToModel(mPoint=mPoint, result=result)
            self.modelQueue.put(mPoint)

            deltaRA = mPoint.mData.raMJNow._degrees - mPoint.mData.raSJNow._degrees
            deltaDEC = mPoint.mData.decMJNow.degrees - mPoint.mData.decSJNow.degrees
            error = np.sqrt(np.square(deltaRA) + np.square(deltaDEC))

            text = f'Solved   image-{mPoint.mParam.count:03d} ->   '
            text += f'Ra: {transform.convertToHMS(result.solve.raJ2000)} '
            text += f'({result.solve.raJ2000.hours:4.3f}), '
            text += f'Dec: {transform.convertToDMS(result.solve.decJ2000)} '
            text += f'({result.solve.decJ2000.degrees:4.3f}), '
            text += f'Error: {error:5.1f}, '
            text += f'Angle: {result.solve.angle:3.0f}, '
            text += f'Scale: {result.solve.scale:4.3f}'
            self.app.message.emit(text, 0)
        else:
            text = f'Solving error for image-{mPoint.mParam.count:03d}'
            self.app.message.emit(text, 2)

        self.updateProgress(number=number, count=count, modelingDone=modelingDone)

        if modelingDone:
            self.modelFinished()

        return True

    def modelSolve(self):
        """
        modelSolve is the method called from the signal image saved and starts the solving
        process for this image. therefore it takes the model point from the queue and uses
        the parameters stored. if the queue is empty (which should be to the case), it
        just returns.

        after starting the solving process in a threaded way (should run in parallel to gui)
        it puts the model point to the next queue, the result queue.

        in addition if the image window is present, it send a signal for displaying the
        actual captured image.

        it shows the actual processed point index in GUI

        :return: success
        """

        if self.solveQueue.empty():
            self.logger.error('empty solve queue')
            return False

        mPoint = self.solveQueue.get()

        # showing the expose image in the image window
        if self.app.imageW:
            self.app.imageW.signals.showImage.emit(mPoint.mParam.path)

        self.app.astrometry.solveThreading(app=mPoint.mParam.astrometry,
                                           fitsPath=mPoint.mParam.path,
                                           radius=mPoint.mParam.radius,
                                           timeout=mPoint.mParam.timeout,
                                           updateFits=False,
                                           )
        self.resultQueue.put(mPoint)

        text = f'Solving  image-{mPoint.mParam.count:03d} ->   '
        text += f'{os.path.basename(mPoint.mParam.path)}'
        self.app.message.emit(text, 0)
        self.ui.mSolve.setText(f'{mPoint.mParam.count + 1:2d}')

        return True

    def modelImage(self):
        """
        modelImage is the method called from the signal mount and dome slewed finish and
        starts the imaging for the model point. therefore it takes the model point from
        the queue and uses the parameters stored. if the queue is empty (which should be to
        the case), it just returns.

        as we are combining the reception of multiple signals for detecting that all slew
        actions are finished, we have to reset the collector Class for preparing a new
        cycle.

        after the imaging with parameters started, the actual mount data (coordinates,
        time, pierside) is added to the model point as this information is later needed for
        solving and building the model itself.

        it shows the actual processed point index in GUI

        :return: success
        """

        if self.imageQueue.empty():
            self.logger.error('empty image queue')
            return False

        mPoint = self.imageQueue.get()
        self.collector.resetSignals()

        self.app.imaging.expose(imagePath=mPoint.mParam.path,
                                expTime=mPoint.iParam.expTime,
                                binning=mPoint.iParam.binning,
                                subFrame=mPoint.iParam.subFrame,
                                fastReadout=mPoint.iParam.fastReadout,
                                )

        mPoint = MPoint(mParam=mPoint.mParam,
                        iParam=mPoint.iParam,
                        point=mPoint.point,
                        mData=MData(raMJNow=self.app.mount.obsSite.raJNow,
                                    decMJNow=self.app.mount.obsSite.decJNow,
                                    raSJNow=None,
                                    decSJNow=None,
                                    sidereal=self.app.mount.obsSite.timeSidereal,
                                    julian=self.app.mount.obsSite.timeJD,
                                    pierside=self.app.mount.obsSite.pierside,
                                    ),
                        rData=None,
                        )

        self.solveQueue.put(mPoint)

        text = f'Exposing image-{mPoint.mParam.count:03d} ->   '
        text += f'path: {os.path.basename(mPoint.mParam.path)}'
        self.app.message.emit(text, 0)
        self.ui.mImage.setText(f'{mPoint.mParam.count + 1 :2d}')

        return True

    def modelSlew(self):
        """
        modelSlew is the method called from the model core method and is the beginning of
        the modeling chain. it starts with taking a first model point from the initial
        slew queue and starts slewing mount (and dome if present).if the queue is empty
        (which should be to the case), it just returns.

        it shows the actual processed point index in GUI

        :return: success

        """

        if self.slewQueue.empty():
            return False

        mPoint = self.slewQueue.get()

        self.app.dome.slewToAltAz(azimuthMount=mPoint.point.azimuth)
        self.app.mount.obsSite.slewAltAz(alt_degrees=mPoint.point.altitude,
                                         az_degrees=mPoint.point.azimuth,
                                         )

        self.imageQueue.put(mPoint)

        text = f'Slewing  image-{mPoint.mParam.count:03d} ->   '
        text += f'altitude: {mPoint.point.altitude:3.0f}, '
        text += f'azimuth: {mPoint.point.azimuth:3.0f}'
        self.app.message.emit(text, 0)
        self.ui.mPoints.setText(f'{mPoint.mParam.number:2d}')
        self.ui.mSlew.setText(f'{mPoint.mParam.count + 1:2d}')

        return True

    def clearQueues(self):
        """
        clearQueues ensures that all used queues will be emptied before starting

        :return: true for test purpose
        """

        self.slewQueue.queue.clear()
        self.imageQueue.queue.clear()
        self.solveQueue.queue.clear()
        self.resultQueue.queue.clear()
        self.modelQueue.queue.clear()

        return True

    def prepareGUI(self):
        """
        prepareGUI sets GUI elements to state, whereas there will be no influence for
        running actions

        :return: true for test purpose
        """

        self.ui.batchModel.setEnabled(False)

        return True

    def defaultGUI(self):
        """
        defaultGUI will reset all gui elements to the idle or default state and new actions
        could be started again

        :return: true for test purpose
        """

        self.changeStyleDynamic(self.ui.runFullModel, 'running', False)
        self.changeStyleDynamic(self.ui.cancelFullModel, 'cancel', False)
        self.changeStyleDynamic(self.ui.runAlignModel, 'running', False)
        self.changeStyleDynamic(self.ui.cancelAlignModel, 'cancel', False)
        self.ui.runFullModel.setEnabled(True)
        self.ui.cancelFullModel.setEnabled(False)
        self.ui.runAlignModel.setEnabled(True)
        self.ui.cancelAlignModel.setEnabled(False)
        self.ui.batchModel.setEnabled(True)
        self.ui.plateSolveSync.setEnabled(True)
        self.ui.runFlexure.setEnabled(True)
        self.ui.runHysteresis.setEnabled(True)
        self.ui.timeToFinish.setText('00:00')
        self.ui.timeElapsed.setText('00:00')
        self.ui.timeFinished.setText('00:00:00')
        self.ui.mPoints.setText('0')
        self.ui.mSlew.setText('0')
        self.ui.mImage.setText('0')
        self.ui.mSolve.setText('0')
        self.ui.modelProgress.setValue(0)

        return True

    def prepareSignals(self):
        """
        prepareSignals establishes the signals chain. as we have multiple actions running
        at the same time, the synchronisation by the right link of the signals.

        first we link the two slew finished signals to modelImage. that means as soon as
        both slew finished signals are received, the imaging will be started
        when download of an image starts, we could slew to another point
        when image is saved, we could start with solving

        :return: true for test purpose
        """

        self.collector.addWaitableSignal(self.app.mount.signals.slewFinished)
        if self.app.dome.device is not None:
            self.collector.addWaitableSignal(self.app.dome.signals.slewFinished)

        self.collector.ready.connect(self.modelImage)
        self.app.imaging.signals.integrated.connect(self.modelSlew)
        self.app.imaging.signals.saved.connect(self.modelSolve)
        self.app.astrometry.signals.done.connect(self.modelSolveDone)

        return True

    def defaultSignals(self):
        """
        defaultSignals clears the signal queue and removes the connections

        :return: true for test purpose
        """

        self.app.imaging.signals.saved.disconnect(self.modelSolve)
        self.app.imaging.signals.integrated.disconnect(self.modelSlew)
        self.app.astrometry.signals.done.disconnect(self.modelSolveDone)
        self.collector.ready.disconnect(self.modelImage)
        self.collector.clear()

        return True

    def cancelFull(self):
        """
        cancelFull aborts imaging and stops all modeling queues and actions

        :return: true for test purpose
        """

        self.app.imaging.abort()
        self.app.astrometry.abort()
        self.defaultSignals()
        self.clearQueues()
        self.defaultGUI()
        self.app.message.emit('Modeling cancelled', 2)

        return True

    def retrofitModel(self, model=None):
        """
        retrofitModel reads the actual model points and results out of the mount computer
        and adds the optimized (recalculated) error values to the point. that's necessary,
        because when imaging and solving a point the error is related to this old model.
        when programming a new model, all point will be recalculated be the mount
        computer an get a new error value which is based on the new model.

        :param model:
        :return: updated model
        """

        if model is None:
            self.logger.debug('model is None')
            return list()

        starList = self.app.mount.model.starList

        if len(starList) != len(model):
            text = f'length starList [{len(starList)}] and length '
            text += f'model [{len(model)}] is different'
            self.logger.debug(text)
            return list()

        for i, mPoint in enumerate(model):
            rData = RData(errorRMS=starList[i].errorRMS,
                          errorRA=starList[i].errorRA(),
                          errorDEC=starList[i].errorDEC(),
                          )
            mPoint = MPoint(mParam=mPoint.mParam,
                            iParam=mPoint.iParam,
                            point=mPoint.point,
                            mData=mPoint.mData,
                            rData=rData,
                            )
            model[i] = mPoint

        return model

    @staticmethod
    def generateSaveModel(model=None):
        """
        generateSaveModel builds from the model file a format which could be serialized
        in json. this format will be used for storing model on file

        :param model:
        :return: save model format
        """

        saveModel = list()
        for mPoint in model:
            sPoint = {'name': mPoint.mParam.name,
                      'path': mPoint.mParam.path,
                      'number': mPoint.mParam.number,
                      'count': mPoint.mParam.count,
                      'expTime': mPoint.iParam.expTime,
                      'binning': mPoint.iParam.binning,
                      'subFrame': mPoint.iParam.subFrame,
                      'fastReadout': mPoint.iParam.fastReadout,
                      'altitude': mPoint.point.altitude,
                      'azimuth': mPoint.point.azimuth,
                      'raMJNow': mPoint.mData.raMJNow.hours,
                      'decMJNow': mPoint.mData.decMJNow.degrees,
                      'raSJNow': mPoint.mData.raSJNow.hours,
                      'decSJNow': mPoint.mData.decSJNow.degrees,
                      'sidereal': mPoint.mData.sidereal.hours,
                      'julian': mPoint.mData.julian.utc_iso(),
                      'pierside': mPoint.mData.pierside,
                      'errorRMS': mPoint.rData.errorRMS,
                      'errorRa': mPoint.rData.errorRA,
                      'errorDEC': mPoint.rData.errorDEC,
                      }
            saveModel.append(sPoint)
        return saveModel

    def saveModel(self, model=None):
        """
        saveModel saves the model data for later use. with this data, the model could
        be reprogrammed without doing some imaging, it could be added with other data to
        extend the model to a broader base.
        in addition it should be possible to make som analyses with this data.

        :param model:
        :return: success
        """

        if model is None:
            return False
        if len(model) < 3:
            self.logger.debug(f'only {len(model)} points available')
            return False

        saveModel = self.generateSaveModel(model)

        self.app.message.emit(f'writing model [{self.modelName}]', 0)

        modelPath = f'{self.app.mwGlob["modelDir"]}/{self.modelName}.model'
        with open(modelPath, 'w') as outfile:
            json.dump(saveModel,
                      outfile,
                      sort_keys=True,
                      indent=4)
        return True

    def collectModelData(self):
        """
        collectModelData writes all model point from the queue to a data structure for
        later use.

        :return: model
        """

        model = list()
        while not self.modelQueue.empty():
            mPoint = self.modelQueue.get()
            model.append(mPoint)

        return model

    @staticmethod
    def generateBuildDataFromJSON(loadModel=None):
        """
        generateBuildData takes the model data and generates from it a data structure
        needed for programming the model into the mount computer.

        :param loadModel:
        :return: build
        """

        build = list()

        for mPoint in loadModel:
            # prepare data
            mCoord = (mPoint['raMJNow'], mPoint['decMJNow'])
            sCoord = (mPoint['raSJNow'], mPoint['decSJNow'])
            sidereal = mPoint['sidereal']
            pierside = mPoint['pierside']

            # combine data into structure
            programmingPoint = APoint(mCoord=mCoord,
                                      sCoord=sCoord,
                                      sidereal=sidereal,
                                      pierside=pierside,
                                      )
            build.append(programmingPoint)

        return build

    @staticmethod
    def generateBuildData(model=None):
        """
        generateBuildData takes the model data and generates from it a data structure
        needed for programming the model into the mount computer.

        :param model:
        :return: build
        """

        build = list()

        for mPoint in model:
            # prepare data
            mCoord = skyfield.api.Star(ra=mPoint.mData.raMJNow,
                                       dec=mPoint.mData.decMJNow)
            sCoord = skyfield.api.Star(ra=mPoint.mData.raSJNow,
                                       dec=mPoint.mData.decSJNow)
            sidereal = mPoint.mData.sidereal
            pierside = mPoint.mData.pierside

            # combine data into structure
            programmingPoint = APoint(mCoord=mCoord,
                                      sCoord=sCoord,
                                      sidereal=sidereal,
                                      pierside=pierside,
                                      )
            build.append(programmingPoint)

        return build

    def modelFinished(self):
        """
        modelFinished is called when tha last point was solved. in addition the saving
        of the model and the programming is done here.

        is the flag delete images after modeling is set, the entire directory will be
        deleted

        :return: true for test purpose
        """

        model = self.collectModelData()
        build = self.generateBuildData(model=model)

        # stopping other activities
        self.defaultSignals()
        self.clearQueues()
        self.defaultGUI()

        # finally do it
        self.app.message.emit('Programming model to mount', 0)
        suc = self.app.mount.model.programAlign(build)

        if suc:
            self.app.message.emit('Model programmed with success', 0)
            self.refreshModel()
            model = self.retrofitModel(model=model)
            self.saveModel(model=model)
        else:
            self.app.message.emit('Model programming error', 2)

        # cleaning up the disk space
        if not self.ui.checkKeepImages.isChecked():
            self.app.message.emit('Deleting model images', 0)
            shutil.rmtree(self.imageDir, ignore_errors=True)

        self.app.message.emit('Modeling finished', 1)
        self.playAudioModelFinished()

        return True

    def modelCore(self, points=None):
        """
        modelCore is the main method for preparing a model run. in addition it checks
        necessary components and prepares all the parameters.
        the modeling queue will be filled with point and the queue is started.

        :param points:
        :return: true for test purpose
        """

        if not points:
            return False

        app = self.app.mainW.ui.astrometryDevice.currentText()
        if app.startswith('No device'):
            return False

        # collection locations for files
        nameTime = self.app.mount.obsSite.timeJD.utc_strftime('%Y-%m-%d-%H-%M-%S')
        self.modelName = f'm-{self.lastModelType}-{nameTime}'
        self.imageDir = f'{self.app.mwGlob["imageDir"]}/{self.modelName}'
        if not os.path.isdir(self.imageDir):
            os.mkdir(self.imageDir)
        if not os.path.isdir(self.imageDir):
            return False

        self.clearQueues()
        self.app.message.emit(f'Modeling {self.modelName} started', 1)

        # collection all necessary information
        expTime = self.app.mainW.ui.expTime.value()
        binning = self.app.mainW.ui.binning.value()
        subFrame = self.app.mainW.ui.subFrame.value()
        fastReadout = self.app.mainW.ui.checkFastDownload.isChecked()
        solveTimeout = self.ui.solveTimeout.value()
        searchRadius = self.ui.searchRadius.value()

        # setting overall parameters
        settleMount = self.app.mainW.ui.settleTimeMount.value()
        settleDome = self.app.mainW.ui.settleTimeDome.value()
        self.app.mount.settlingTime = settleMount
        self.app.dome.settlingTime = settleDome

        self.prepareGUI()
        self.prepareSignals()
        self.startModeling = time.time()

        # queuing modeling points
        number = len(points)
        for count, point in enumerate(points):
            # define the path to the image file
            path = f'{self.imageDir}/image-{count:03d}.fits'

            iParam = IParam(expTime=expTime,
                            binning=binning,
                            subFrame=subFrame,
                            fastReadout=fastReadout,
                            )
            mParam = MParam(number=number,
                            count=count,
                            path=path,
                            name=self.modelName,
                            astrometry=app,
                            timeout=solveTimeout,
                            radius=searchRadius,
                            )

            # transfer to working in a queue with necessary data
            self.slewQueue.put(MPoint(iParam=iParam,
                                      mParam=mParam,
                                      point=Point(altitude=point[0],
                                                  azimuth=point[1]),
                                      mData=None,
                                      rData=None,
                                      )
                               )
        # kick off modeling
        self.modelSlew()

        return True

    def modelFull(self):
        """
        modelFull sets the adequate gui elements, selects the model points and calls the
        core modeling method

        :return: true for test purpose
        """

        # checking constraints for modeling
        points = self.app.data.buildP
        number = len(points)
        if not 2 < number < 100:
            return False

        self.changeStyleDynamic(self.ui.runFullModel, 'running', True)
        self.changeStyleDynamic(self.ui.cancelFullModel, 'cancel', True)
        self.ui.cancelFullModel.setEnabled(True)
        self.ui.runAlignModel.setEnabled(False)
        self.ui.plateSolveSync.setEnabled(False)
        self.ui.runFlexure.setEnabled(False)
        self.ui.runHysteresis.setEnabled(False)
        self.ui.cancelFullModel.setEnabled(True)

        suc = self.modelCore(points=points)
        if not suc:
            self.defaultGUI()
            return False

        return True

    def modelAlign(self):
        """
        modelAlign sets the adequate gui elements, selects the model points and calls the
        core modeling method

        :return: true for test purpose
        """

        # checking constraints for modeling
        points = self.app.data.buildP
        number = len(points)
        if not 2 < number < 100:
            return False

        self.changeStyleDynamic(self.ui.runAlignModel, 'running', True)
        self.changeStyleDynamic(self.ui.cancelAlignModel, 'cancel', True)
        self.ui.cancelAlignModel.setEnabled(True)
        self.ui.runFullModel.setEnabled(False)
        self.ui.plateSolveSync.setEnabled(False)
        self.ui.runFlexure.setEnabled(False)
        self.ui.runHysteresis.setEnabled(False)

        suc = self.modelCore(points=points)
        if not suc:
            self.defaultGUI()
            return False

        return True

    def loadProgramModel(self):
        """
        loadProgramModel selects one or more models from the files system, combines them
        if more than one was selected and programs them into the mount computer.

        :return: success
        """

        folder = self.app.mwGlob['modelDir']
        loadFilePath, fileName, ext = self.openFile(self,
                                                    'Open model file',
                                                    folder,
                                                    'Model files (*.model)',
                                                    multiple=True,
                                                    )
        if not loadFilePath:
            return False

        if isinstance(loadFilePath, str):
            loadFilePath = [loadFilePath]

        self.app.message.emit('Programing stored models', 1)
        modelJSON = list()
        for file in loadFilePath:
            self.app.message.emit(f'Using model [{file}]', 0)
            with open(file, 'r') as infile:
                model = json.load(infile)
                modelJSON += model

        if len(modelJSON) > 99:
            self.app.message.emit('Combined models have more than 99 points', 2)
            return False

        build = self.generateBuildDataFromJSON(modelJSON)

        # finally do it
        self.app.message.emit('Programming model to mount', 0)
        suc = self.app.mount.model.programAlign(build)

        if suc:
            self.app.message.emit('Model programmed with success', 0)
            self.refreshModel()
        else:
            self.app.message.emit('Model programming error', 2)

        return suc



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
# Python  v3.7.4
#
# Michael Würtenberger
# (c) 2019
#
# Licence APL2.0
#
###########################################################
# standard libraries
import logging
import zlib
import os
from datetime import datetime
# external packages
import PyQt5
import numpy as np
import astropy.io.fits as fits
# local imports
from mw4.base import indiClass


class CameraSignals(PyQt5.QtCore.QObject):
    """
    The CameraSignals class offers a list of signals to be used and instantiated by
    the Mount class to get signals for triggers for finished tasks to
    enable a gui to update their values transferred to the caller back.

    This has to be done in a separate class as the signals have to be subclassed from
    QObject and the Mount class itself is subclassed from object
    """

    __all__ = ['CameraSignals']
    version = '0.100.0'

    integrated = PyQt5.QtCore.pyqtSignal()
    saved = PyQt5.QtCore.pyqtSignal(object)
    message = PyQt5.QtCore.pyqtSignal(object)


class Camera(indiClass.IndiClass):
    """
    the class Camera inherits all information and handling of the Camera device.


        >>> fw = Camera(
        >>>           app=app
        >>>           host=host
        >>>           name=''
        >>>          )
    """

    __all__ = ['Camera',
               ]

    version = '0.100.0'
    logger = logging.getLogger(__name__)

    # update rate to 1000 milli seconds for setting indi server
    UPDATE_RATE = 1000

    def __init__(self,
                 app=None,
                 host=None,
                 name='',
                 ):
        super().__init__(host=host,
                         name=name
                         )

        self.app = app
        self.signals = CameraSignals()
        self.imagePath = ''

    def setUpdateConfig(self, deviceName):
        """
        _setUpdateRate corrects the update rate of camera devices to get an defined
        setting regardless, what is setup in server side.

        :param deviceName:
        :return: success
        """

        if deviceName != self.name:
            return False

        if self.device is None:
            return False

        # set BLOB mode also
        self.client.setBlobMode(blobHandling='Also',
                                deviceName=deviceName)
        # setting a object name
        objectName = self.device.getText('FITS_HEADER')
        objectName['FITS_OBJECT'] = 'skyview'
        self.client.sendNewText(deviceName=deviceName,
                                propertyName='FITS_HEADER',
                                elements=objectName,
                                )
        # setting WCS Control off
        wcs = self.device.getSwitch('WCS_CONTROL')
        wcs['WCS_DISABLE'] = True
        self.client.sendNewSwitch(deviceName=deviceName,
                                  propertyName='WCS_CONTROL',
                                  elements=wcs,
                                  )
        # setting active device for telescope
        telescope = self.device.getText('ACTIVE_DEVICES')
        telescope['ACTIVE_TELESCOPE'] = 'LX200 10micron'
        self.client.sendNewText(deviceName=deviceName,
                                propertyName='ACTIVE_DEVICES',
                                elements=telescope,
                                )
        # setting polling updates in driver
        update = self.device.getNumber('POLLING_PERIOD')
        if 'PERIOD_MS' not in update:
            return False
        if update.get('PERIOD_MS', 0) == self.UPDATE_RATE:
            return True
        update['PERIOD_MS'] = self.UPDATE_RATE
        suc = self.client.sendNewNumber(deviceName=deviceName,
                                        propertyName='POLLING_PERIOD',
                                        elements=update,
                                        )

        return suc

    def setPixelSize(self, propertyName='', element='', value=0):
        """

        :param propertyName:
        :param element:
        :param value:
        :return: success
        """

        if propertyName == 'CCD_INFO':
            if element == 'CCD_PIXEL_SIZE_X':
                self.app.mainW.ui.pixelSize.setValue(np.round(value, 1))
                return True
        return False

    def setExposureState(self, propertyName='', value=0):
        """

        :param propertyName:
        :param value:
        :return: success
        """

        if propertyName == 'CCD_EXPOSURE':
            if not hasattr(self.device, 'CCD_EXPOSURE'):
                return False
            if self.device.CCD_EXPOSURE['state'] == 'Idle':
                self.signals.message.emit('')
            elif self.device.CCD_EXPOSURE['state'] == 'Busy':
                if value == 0:
                    self.signals.integrated.emit()
                    self.signals.message.emit('download')
                else:
                    self.signals.message.emit(f'expose {value:2.0f} s')
            elif self.device.CCD_EXPOSURE['state'] == 'Ok':
                self.signals.message.emit('')
            return True
        else:
            return False

    def updateNumber(self, deviceName, propertyName):
        """
        updateNumber is called whenever a new number is received in client. it runs
        through the device list and writes the number data to the according locations.

        :param deviceName:
        :param propertyName:
        :return:
        """

        if self.device is None:
            return False
        if deviceName != self.name:
            return False

        for element, value in self.device.getNumber(propertyName).items():
            key = propertyName + '.' + element
            self.data[key] = value
            # print(propertyName, element, value)

            self.setPixelSize(propertyName=propertyName, element=element, value=value)
            self.setExposureState(propertyName=propertyName, value=value)

        return True

    def updateText(self, deviceName, propertyName):
        """
        updateNumber is called whenever a new number is received in client. it runs
        through the device list and writes the number data to the according locations.

        :param deviceName:
        :param propertyName:
        :return:
        """

        if self.device is None:
            return False
        if deviceName != self.name:
            return False

        for element, value in self.device.getText(propertyName).items():
            key = propertyName + '.' + element
            self.data[key] = value
            # print(propertyName, element, value)
        return True

    def updateSwitch(self, deviceName, propertyName):
        """
        updateNumber is called whenever a new number is received in client. it runs
        through the device list and writes the number data to the according locations.

        :param deviceName:
        :param propertyName:
        :return:
        """

        if self.device is None:
            return False
        if deviceName != self.name:
            return False

        for element, value in self.device.getSwitch(propertyName).items():
            key = propertyName + '.' + element
            self.data[key] = value
            # print(propertyName, element, value)
        return True

    def updateLight(self, deviceName, propertyName):
        """
        updateNumber is called whenever a new number is received in client. it runs
        through the device list and writes the number data to the according locations.

        :param deviceName:
        :param propertyName:
        :return:
        """

        if self.device is None:
            return False
        if deviceName != self.name:
            return False

        for element, value in self.device.getLight(propertyName).items():
            key = propertyName + '.' + element
            self.data[key] = value
            # print(propertyName, element, value)
        return True

    def updateBLOB(self, deviceName, propertyName):
        """
        updateBLOB is called whenever a new BLOB is received in client. it runs
        through the device list and writes the number data to the according locations.

        :param deviceName:
        :param propertyName:
        :return:
        """

        if self.device is None:
            return False
        if deviceName != self.name:
            return False

        data = self.device.getBlob(propertyName)

        if 'value' not in data:
            return False
        if data['name'] != 'CCD1':
            return False
        if not self.imagePath:
            return False
        if not os.path.isdir(os.path.dirname(self.imagePath)):
            return False

        if data['format'] == '.fits.fz':
            HDU = fits.HDUList.fromstring(data['value'])
            fits.writeto(self.imagePath, HDU[0].data, HDU[0].header, overwrite=True)
            self.logger.debug('Image BLOB is in FPacked format')

        elif data['format'] == '.fits.z':
            HDU = fits.HDUList.fromstring(zlib.decompress(data['value']))
            fits.writeto(self.imagePath, HDU[0].data, HDU[0].header, overwrite=True)
            self.logger.debug('Image BLOB is compressed fits format')

        elif data['format'] == '.fits':
            HDU = fits.HDUList.fromstring(data['value'])
            fits.writeto(self.imagePath, HDU[0].data, HDU[0].header, overwrite=True)
            self.logger.debug('Image BLOB is uncompressed fits format')

        else:
            self.logger.debug('Image BLOB is not supported')

        self.signals.saved.emit(self.imagePath)
        return True

    def canSubFrame(self, subFrame=100):
        """
        canSubFrame checks if a camera supports sub framing and reports back

        :param subFrame:
        :return: success
        """
        if subFrame > 100:
            return False
        if subFrame < 10:
            return False
        if 'CCD_FRAME.X' not in self.data or 'CCD_FRAME.Y' not in self.data:
            return False

        return True

    def canBinning(self, binning=1):
        """
        canBinning checks if the camera supports that type of binning

        :param binning:
        :return: success
        """
        if binning < 1:
            return False
        if binning > 4:
            return False
        if 'CCD_BINNING.HOR_BIN' not in self.data:
            return False

        return True

    def calcSubFrame(self, subFrame=100):
        """
        calcSubFrame calculates the subFrame parameters depending on the percentage of
        the reduction. the subFrame will be centered on the image area.

        :param subFrame: percentage 0-100 of
        :return:
        """
        if subFrame < 10 or subFrame > 100:
            width = self.data['CCD_INFO.CCD_MAX_X']
            height = self.data['CCD_INFO.CCD_MAX_Y']
            posX = 0
            posY = 0
        else:
            width = int(self.data['CCD_INFO.CCD_MAX_X'] * subFrame / 100)
            height = int(self.data['CCD_INFO.CCD_MAX_Y'] * subFrame / 100)
            posX = int((self.data['CCD_INFO.CCD_MAX_X'] - width) / 2)
            posY = int((self.data['CCD_INFO.CCD_MAX_Y'] - height) / 2)

        return posX, posY, width, height

    def setupFrameCompress(self):
        """
        setupFrameCompress prepares the overall INDI setup data for imaging

        :return: success
        """

        # setting compression to on as default
        indiCmd = self.device.getSwitch('CCD_COMPRESSION')
        if 'CCD_COMPRESS' not in indiCmd:
            return False
        indiCmd['CCD_COMPRESS'] = True
        suc = self.client.sendNewSwitch(deviceName=self.name,
                                        propertyName='CCD_COMPRESSION',
                                        elements=indiCmd,
                                        )
        if not suc:
            return False

        # setting frame type to light
        indiCmd = self.device.getSwitch('CCD_FRAME_TYPE')
        if 'FRAME_LIGHT' not in indiCmd:
            return False
        indiCmd['FRAME_LIGHT'] = True
        suc = self.client.sendNewSwitch(deviceName=self.name,
                                        propertyName='CCD_FRAME_TYPE',
                                        elements=indiCmd,
                                        )
        return suc

    def expose(self, imagePath='', expTime=3, binning=1,
               subFrame=100, fastReadout=True):
        """

        :param imagePath:
        :param expTime:
        :param binning:
        :param subFrame:
        :param fastReadout:
        :return: success
        """

        if not imagePath:
            return False
        if not self.canSubFrame(subFrame=subFrame):
            return False
        if not self.canBinning(binning=binning):
            return False

        self.imagePath = imagePath

        suc = self.setupFrameCompress()
        if not suc:
            return False

        # setting binning value for x and y equally
        indiCmd = self.device.getNumber('CCD_BINNING')
        indiCmd['HOR_BIN'] = binning
        indiCmd['VER_BIN'] = binning
        suc = self.client.sendNewNumber(deviceName=self.name,
                                        propertyName='CCD_BINNING',
                                        elements=indiCmd,
                                        )
        if not suc:
            return False

        # setting subFrame
        posX, posY, width, height = self.calcSubFrame(subFrame)

        indiCmd = self.device.getNumber('CCD_FRAME')
        indiCmd['X'] = posX
        indiCmd['Y'] = posY
        indiCmd['WIDTH'] = width
        indiCmd['HEIGHT'] = height
        suc = self.client.sendNewNumber(deviceName=self.name,
                                        propertyName='CCD_FRAME',
                                        elements=indiCmd,
                                        )
        if not suc:
            return False

        # setting fast mode:
        quality = self.device.getSwitch('READOUT_QUALITY')
        self.logger.debug(f'camera has readout quality entry: {quality}')
        quality['QUALITY_LOW'] = fastReadout
        quality['QUALITY_HIGH'] = not fastReadout
        self.client.sendNewSwitch(deviceName=self.name,
                                  propertyName='READOUT_QUALITY',
                                  elements=quality,
                                  )

        # setting and starting exposure
        indiCmd = self.device.getNumber('CCD_EXPOSURE')
        indiCmd['CCD_EXPOSURE_VALUE'] = expTime
        suc = self.client.sendNewNumber(deviceName=self.name,
                                        propertyName='CCD_EXPOSURE',
                                        elements=indiCmd,
                                        )
        return suc

    def abort(self):
        """
        abort cancels the exposing

        :return: success
        """

        indiCmd = self.device.getSwitch('CCD_ABORT_EXPOSURE')
        if 'ABORT' not in indiCmd:
            return False
        indiCmd['ABORT'] = True
        suc = self.client.sendNewSwitch(deviceName=self.name,
                                        propertyName='CCD_ABORT_EXPOSURE',
                                        elements=indiCmd,
                                        )

        return suc
